from typing import List

from drf_renderer_xlsx.mixins import XLSXFileMixin
from rest_framework import status
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet, ReadOnlyModelViewSet

from NEMO.models import (
	Account,
	Area,
	AreaAccessRecord,
	Project,
	Reservation,
	Resource,
	ScheduledOutage,
	StaffCharge,
	Task,
	Tool,
	TrainingSession,
	UsageEvent,
	User,
)
from NEMO.serializers import (
	AccountSerializer,
	AreaAccessRecordSerializer,
	AreaSerializer,
	BillableItemSerializer,
	ProjectSerializer,
	ReservationSerializer,
	ResourceSerializer,
	ScheduledOutageSerializer,
	StaffChargeSerializer,
	TaskSerializer,
	ToolSerializer,
	TrainingSessionSerializer,
	UsageEventSerializer,
	UserSerializer,
)
from NEMO.utilities import export_format_datetime
from NEMO.views.api_billing import (
	BillableItem,
	BillingFilterForm,
	get_area_access_for_billing,
	get_consumables_for_billing,
	get_missed_reservations_for_billing,
	get_staff_charges_for_billing,
	get_training_sessions_for_billing,
	get_usage_events_for_billing,
)


class UserViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = User.objects.all()
	serializer_class = UserSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"type": ["exact", "in"],
		"domain": ["exact", "in", "isempty"],
		"username": ["exact", "in"],
		"first_name": ["exact", "icontains"],
		"last_name": ["exact", "icontains"],
		"email": ["exact", "icontains"],
		"badge_number": ["exact"],
		"is_active": ["exact"],
		"is_staff": ["exact"],
		"is_facility_manager": ["exact"],
		"is_superuser": ["exact"],
		"is_service_personnel": ["exact"],
		"is_technician": ["exact"],
		"training_required": ["exact"],
		"date_joined": ["month", "year", "gte", "gt", "lte", "lt"],
		"last_login": ["month", "year", "gte", "gt", "lte", "lt", "isnull"],
		"access_expiration": ["month", "year", "gte", "gt", "lte", "lt", "isnull"],
	}

	def get_filename(self, *args, **kwargs):
		return f"users-{export_format_datetime()}.xlsx"


class ProjectViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Project.objects.all()
	serializer_class = ProjectSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"name": ["exact"],
		"application_identifier": ["exact"],
		"active": ["exact"],
		"account_id": ["exact", "in"],
	}

	def get_filename(self, *args, **kwargs):
		return f"projects-{export_format_datetime()}.xlsx"


class AccountViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Account.objects.all()
	serializer_class = AccountSerializer
	filterset_fields = {"id": ["exact", "in"], "name": ["exact"], "active": ["exact"]}

	def get_filename(self, *args, **kwargs):
		return f"accounts-{export_format_datetime()}.xlsx"


class ToolViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Tool.objects.all()
	serializer_class = ToolSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"name": ["exact"],
		"visible": ["exact"],
		"_category": ["exact"],
		"_operational": ["exact"],
		"_location": ["exact"],
		"_requires_area_access": ["exact", "isnull"],
		"_post_usage_questions": ["isempty"],
	}

	def get_filename(self, *args, **kwargs):
		return f"tools-{export_format_datetime()}.xlsx"


class AreaViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Area.objects.all()
	serializer_class = AreaSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"name": ["exact"],
		"parent_area": ["exact", "in"],
		"category": ["exact"],
		"requires_reservation": ["exact"],
		"buddy_system_allowed": ["exact"],
		"maximum_capacity": ["exact", "gte", "gt", "lte", "lt", "isnull"],
		"count_staff_in_occupancy": ["exact"],
		"count_service_personnel_in_occupancy": ["exact"],
	}

	def get_filename(self, *args, **kwargs):
		return f"areas-{export_format_datetime()}.xlsx"


class ResourceViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Resource.objects.all()
	serializer_class = ResourceSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"name": ["exact"],
		"available": ["exact"],
		"fully_dependent_tools": ["in"],
		"partially_dependent_tools": ["in"],
		"dependent_areas": ["in"],
	}

	def get_filename(self, *args, **kwargs):
		return f"resources-{export_format_datetime()}.xlsx"


class ReservationViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Reservation.objects.all()
	serializer_class = ReservationSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"start": ["gte", "gt", "lte", "lt"],
		"end": ["gte", "gt", "lte", "lt", "isnull"],
		"project_id": ["exact", "in"],
		"user_id": ["exact", "in"],
		"creator_id": ["exact", "in"],
		"tool_id": ["exact", "in", "isnull"],
		"area_id": ["exact", "in", "isnull"],
		"cancelled": ["exact"],
		"missed": ["exact"],
		"question_data": ["isempty"],
	}

	def get_filename(self, *args, **kwargs):
		return f"reservations-{export_format_datetime()}.xlsx"


class UsageEventViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = UsageEvent.objects.all()
	serializer_class = UsageEventSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"start": ["gte", "gt", "lte", "lt"],
		"end": ["gte", "gt", "lte", "lt", "isnull"],
		"project_id": ["exact", "in"],
		"user_id": ["exact", "in"],
		"operator_id": ["exact", "in"],
		"tool_id": ["exact", "in"],
	}

	def get_filename(self, *args, **kwargs):
		return f"usage_events-{export_format_datetime()}.xlsx"


class AreaAccessRecordViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = AreaAccessRecord.objects.all().order_by("-start")
	serializer_class = AreaAccessRecordSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"start": ["gte", "gt", "lte", "lt"],
		"end": ["gte", "gt", "lte", "lt", "isnull"],
		"project_id": ["exact", "in"],
		"customer_id": ["exact", "in"],
		"area_id": ["exact", "in"],
		"staff_charge_id": ["exact", "isnull", "in"],
	}

	def get_filename(self, *args, **kwargs):
		return f"area_access_records-{export_format_datetime()}.xlsx"


class TaskViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = Task.objects.all()
	serializer_class = TaskSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"urgency": ["exact", "gte", "gt", "lte", "lt"],
		"tool_id": ["exact", "in"],
		"force_shutdown": ["exact"],
		"safety_hazard": ["exact"],
		"creator_id": ["exact", "in"],
		"creation_time": ["gte", "gt", "lte", "lt"],
		"estimated_resolution_time": ["gte", "gt", "lte", "lt"],
		"problem_category": ["exact", "in"],
		"cancelled": ["exact"],
		"resolved": ["exact"],
		"resolution_time": ["gte", "gt", "lte", "lt"],
		"resolver_id": ["exact", "in"],
		"resolution_category": ["exact", "in"],
	}

	def get_filename(self, *args, **kwargs):
		return f"tasks-{export_format_datetime()}.xlsx"


class ScheduledOutageViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = ScheduledOutage.objects.all()
	serializer_class = ScheduledOutageSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"start": ["gte", "gt", "lte", "lt"],
		"end": ["gte", "gt", "lte", "lt", "isnull"],
		"creator_id": ["exact", "in"],
		"category": ["exact", "in"],
		"tool_id": ["exact", "in", "isnull"],
		"area_id": ["exact", "in", "isnull"],
		"resource_id": ["exact", "in", "isnull"],
	}

	def get_filename(self, *args, **kwargs):
		return f"outages-{export_format_datetime()}.xlsx"


class StaffChargeViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = StaffCharge.objects.all()
	serializer_class = StaffChargeSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"staff_member_id": ["exact", "in"],
		"customer_id": ["exact", "in"],
		"project_id": ["exact", "in"],
		"start": ["gte", "gt", "lte", "lt"],
		"end": ["gte", "gt", "lte", "lt", "isnull"],
		"validated": ["exact"],
		"note": ["contains"],
	}

	def get_filename(self, *args, **kwargs):
		return f"staff_charges-{export_format_datetime()}.xlsx"


class TrainingSessionViewSet(XLSXFileMixin, ReadOnlyModelViewSet):
	queryset = TrainingSession.objects.all()
	serializer_class = TrainingSessionSerializer
	filterset_fields = {
		"id": ["exact", "in"],
		"trainer_id": ["exact", "in"],
		"trainee_id": ["exact", "in"],
		"tool_id": ["exact", "in"],
		"project_id": ["exact", "in"],
		"duration": ["exact", "gte", "lte", "gt", "lt"],
		"type": ["exact", "in"],
		"date": ["gte", "gt", "lte", "lt"],
		"qualified": ["exact"],
	}

	def get_filename(self, *args, **kwargs):
		return f"training_sessions-{export_format_datetime()}.xlsx"


class BillingViewSet(XLSXFileMixin, GenericViewSet):
	serializer_class = BillableItemSerializer

	def list(self, request, *args, **kwargs):
		billing_form = BillingFilterForm(self.request.GET)
		if not billing_form.is_valid():
			return Response(status=status.HTTP_400_BAD_REQUEST, data=billing_form.errors)
		queryset = self.get_queryset()
		serializer = self.serializer_class(queryset, many=True)
		return Response(serializer.data)

	def get_queryset(self):
		billing_form = BillingFilterForm(self.request.GET)
		billing_form.full_clean()

		data: List[BillableItem] = []
		data.extend(get_usage_events_for_billing(billing_form))
		data.extend(get_area_access_for_billing(billing_form))
		data.extend(get_consumables_for_billing(billing_form))
		data.extend(get_missed_reservations_for_billing(billing_form))
		data.extend(get_staff_charges_for_billing(billing_form))
		data.extend(get_training_sessions_for_billing(billing_form))

		data.sort(key=lambda x: x.start, reverse=True)
		return data

	def get_filename(self, *args, **kwargs):
		return f"billing-{export_format_datetime()}.xlsx"



from logging import getLogger
from smtplib import SMTPException
from typing import List

from django.contrib.auth.decorators import login_required
from django.core.validators import validate_email
from django.db.models import Q, QuerySet
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.template import Context, Template
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_POST

from NEMO.decorators import staff_member_required
from NEMO.forms import EmailBroadcastForm
from NEMO.models import Account, Area, Project, Tool, User, UserType
from NEMO.utilities import EmailCategory, send_mail
from NEMO.views.customization import get_customization, get_media_file_contents

logger = getLogger(__name__)


@login_required
@require_GET
def get_email_form(request):
	recipient = request.GET.get('recipient', '')
	try:
		validate_email(recipient)
	except:
		return HttpResponseBadRequest('Recipient not valid.')
	return render(request, 'email/email_form.html', {'recipient': recipient})


@login_required
@require_GET
def get_email_form_for_user(request, user_id):
	recipient = get_object_or_404(User, id=user_id)
	return render(request, 'email/email_form.html', {'name': recipient.get_full_name(), 'recipient': recipient.email})


@login_required
@require_POST
def send_email(request):
	try:
		recipient = request.POST['recipient']
		validate_email(recipient)
		recipient_list = [recipient]
	except:
		return HttpResponseBadRequest('The intended recipient was not a valid email address. The email was not sent.')
	sender = request.user.email
	subject = request.POST.get('subject')
	body = request.POST.get('body')
	if request.POST.get('copy_me'):
		recipient_list.append(sender)
	try:
		send_mail(subject=subject, content=body, from_email=sender, bcc=recipient_list, email_category=EmailCategory.DIRECT_CONTACT)
	except SMTPException as error:
		site_title = get_customization('site_title')
		error_message = f'{site_title} was unable to send the email through the email server. The error message that was received is: ' + str(error)
		logger.exception(error_message)
		dictionary = {
			'title': 'Email not sent',
			'heading': 'There was a problem sending your email',
			'content': error_message,
		}
		return render(request, 'acknowledgement.html', dictionary)
	dictionary = {
		'title': 'Email sent',
		'heading': 'Your email was sent',
	}
	return render(request, 'acknowledgement.html', dictionary)


@staff_member_required
@require_GET
def email_broadcast(request, audience=''):
	dictionary = {}
	if audience == 'tool':
		dictionary['search_base'] = Tool.objects.filter(visible=True)
	elif audience == 'area':
		dictionary['search_base'] = Area.objects.all()
	elif audience == 'project':
		dictionary['search_base'] = Project.objects.filter(active=True, account__active=True)
	elif audience == 'account':
		dictionary['search_base'] = Account.objects.filter(active=True)
	elif audience == 'user':
		user_types = UserType.objects.all()
		dictionary['user_types'] = user_types
		if not user_types:
			return redirect(f"{reverse('compose_email')}?audience={audience}")
	dictionary['audience'] = audience
	return render(request, 'email/email_broadcast.html', dictionary)


@staff_member_required
@require_GET
def compose_email(request):
	try:
		audience = request.GET['audience']
		selection = request.GET.getlist('selection')
		no_type = request.GET.get("no_type") == "on"
		users = get_users_for_email(audience, selection, no_type)
	except:
		dictionary = {'error': 'You specified an invalid audience parameter'}
		return render(request, 'email/email_broadcast.html', dictionary)
	generic_email_sample = get_media_file_contents('generic_email.html')
	dictionary = {
		'audience': audience,
		'selection': selection,
		'no_type': no_type,
		'users': users,
	}
	if generic_email_sample:
		generic_email_context = {
			'title': 'TITLE',
			'greeting': 'Greeting',
			'contents': 'Contents',
			'template_color': '#5bc0de',
		}
		dictionary['generic_email_sample'] = Template(generic_email_sample).render(Context(generic_email_context))
	return render(request, 'email/compose_email.html', dictionary)


@staff_member_required
@require_POST
def send_broadcast_email(request):
	content = get_media_file_contents('generic_email.html')
	if not content:
		return HttpResponseBadRequest('Generic email template not defined. Visit the customization page to upload a template.')
	form = EmailBroadcastForm(request.POST)
	if not form.is_valid():
		return render(request, 'email/compose_email.html', {'form': form})
	dictionary = {
		'title': form.cleaned_data['title'],
		'greeting': form.cleaned_data['greeting'],
		'contents': form.cleaned_data['contents'],
		'template_color': form.cleaned_data['color'],
	}
	content = Template(content).render(Context(dictionary))
	active_choice = form.cleaned_data['only_active_users']
	try:
		audience = form.cleaned_data['audience']
		selection = form.cleaned_data['selection']
		no_type = form.cleaned_data['no_type']
		users = get_users_for_email(audience, selection, no_type)
		if active_choice:
			users = users.filter(is_active=True)
	except Exception as error:
		warning_message = 'Your email was not sent. There was a problem finding the users to send the email to.'
		dictionary = {'error': warning_message}
		logger.warning(warning_message + ' audience: {}, only_active: {}. The error message that was received is: {}'.format(audience, active_choice, str(error)))
		return render(request, 'email/compose_email.html', dictionary)
	if not users:
		dictionary = {'error': 'The audience you specified is empty. You must send the email to at least one person.'}
		return render(request, 'email/compose_email.html', dictionary)
	subject = form.cleaned_data['subject']
	users = [x.email for x in users]
	if form.cleaned_data['copy_me']:
		users += [request.user.email]
	try:
		send_mail(subject=subject, content=content, from_email=request.user.email, bcc=set(users), email_category=EmailCategory.BROADCAST_EMAIL)
	except SMTPException as error:
		site_title = get_customization('site_title')
		error_message = f"{site_title} was unable to send the email through the email server. The error message that was received is: " + str(error)
		logger.exception(error_message)
		dictionary = {
			'title': 'Email not sent',
			'heading': 'There was a problem sending your email',
			'content': error_message,
		}
		return render(request, 'acknowledgement.html', dictionary)
	dictionary = {
		'title': 'Email sent',
		'heading': 'Your email was sent',
	}
	return render(request, 'acknowledgement.html', dictionary)


@staff_member_required
@require_POST
def email_preview(request):
	generic_email_template = get_media_file_contents('generic_email.html')
	if generic_email_template:
		form = EmailBroadcastForm(request.POST)
		email_context = {
			'title': form.data['title'],
			'greeting': form.data['greeting'],
			'contents': form.data['contents'],
			'template_color': form.data['color'],
		}
		email_content = Template(generic_email_template).render(Context(email_context))
		return HttpResponse(mark_safe(email_content))
	return HttpResponse()


def get_users_for_email(audience: str, selection: List, no_type: bool) -> QuerySet:
	users = User.objects.none()
	if audience == 'tool':
		users = User.objects.filter(qualifications__id__in=selection).distinct()
	elif audience == 'area':
		access_levels = Area.objects.get(pk__in=selection).get_physical_access_levels()
		user_filter = Q(physical_access_levels__in=access_levels)
		# if one of the access levels allows staff, add all staff
		if access_levels.filter(allow_staff_access=True).exists():
			user_filter = user_filter | Q(is_staff=True)
		users = User.objects.filter(user_filter).distinct()
	elif audience == 'project':
		users = User.objects.filter(projects__id__in=selection).distinct()
	elif audience == 'account':
		users = User.objects.filter(projects__account__id__in=selection).distinct()
	elif audience == 'user':
		users = User.objects.all().distinct()
		if selection:
			users = users.filter(Q(type_id__in=selection) | Q(type_id__isnull=no_type)) if no_type else users.filter(
				type_id__in=selection)
		elif no_type:
			users = users.filter(type_id__isnull=True)
	return users



from datetime import timedelta

from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.decorators import staff_member_required
from NEMO.models import Reservation, Tool
from NEMO.utilities import localize, naive_local_current_datetime


@staff_member_required
@require_GET
def configuration_agenda(request, time_period='today'):
	tools = Tool.objects.exclude(configuration__isnull=True).exclude(configuration__exclude_from_configuration_agenda=True).values_list('id', flat=True)
	start = None
	end = None
	if time_period == 'today':
		start = naive_local_current_datetime().replace(tzinfo=None)
		end = naive_local_current_datetime().replace(hour=23, minute=59, second=59, microsecond=999, tzinfo=None)
	elif time_period == 'near_future':
		start = naive_local_current_datetime().replace(hour=23, minute=59, second=59, microsecond=999, tzinfo=None)
		if start.weekday() == 4:  # If it's Friday, then the 'near future' is Saturday, Sunday, and Monday
			end = start + timedelta(days=3)
		else:  # If it's not Friday, then the 'near future' is the next day
			end = start + timedelta(days=1)
	start = localize(start)
	end = localize(end)
	reservations = Reservation.objects.filter(start__gt=start, start__lt=end, tool__id__in=tools, self_configuration=False, cancelled=False, missed=False, shortened=False).exclude(additional_information='').order_by('start')
	tools = Tool.objects.filter(id__in=reservations.values_list('tool', flat=True))
	configuration_widgets = {}
	for tool in tools:
		configuration_widgets[tool.id] = tool.configuration_widget(request.user)
	dictionary = {
		'time_period': time_period,
		'tools': tools,
		'reservations': reservations,
		'configuration_widgets': configuration_widgets
	}
	return render(request, 'configuration_agenda.html', dictionary)



# Generated by Django 2.2.13 on 2020-12-10 22:06

from django.conf import settings
from django.db import migrations, models
import django.utils.timezone

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0025_version_3_6_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.7.0", "")

    operations = [
        migrations.CreateModel(
            name='BuddyRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(default=django.utils.timezone.now, help_text='The date and time when the request was created.')),
                ('start', models.DateField(help_text='The start date the user is requesting a buddy.')),
                ('end', models.DateField(help_text='The end date the user is requesting a buddy.')),
                ('description', models.TextField(help_text='The description of the request.')),
                ('expired', models.BooleanField(default=False, help_text="Indicates the request has expired and won't be shown anymore.")),
                ('deleted', models.BooleanField(default=False, help_text="Indicates the request has been deleted and won't be shown anymore.")),
                ('area', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Area')),
                ('user', models.ForeignKey(help_text='The user who is submitting the request.', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='area',
            name='buddy_system_allowed',
            field=models.BooleanField(default=False, help_text='Check this box if the buddy system is allowed in this area.'),
        ),
        migrations.AddField(
            model_name='userpreferences',
            name='display_new_buddy_request_notification',
            field=models.BooleanField(default=True, help_text='Whether or not to notify the user of new buddy requests (via unread badges)', verbose_name='new_buddy_request_notification'),
        ),
        migrations.AddField(
            model_name='userpreferences',
            name='display_new_buddy_request_reply_notification',
            field=models.BooleanField(default=True, help_text='Whether or not to notify the user of replies on buddy request he commented on (via unread badges)', verbose_name='new_buddy_request_reply_notification'),
        ),
        migrations.AddField(
            model_name='userpreferences',
            name='email_new_buddy_request_reply',
            field=models.BooleanField(default=True, help_text='Whether or not to email the user of replies on buddy request he commented on', verbose_name='email_new_buddy_request_reply'),
        ),
        migrations.AlterField(
            model_name='landingpagechoice',
            name='notifications',
            field=models.CharField(blank=True, choices=[('news', 'News creation and updates - notifies all users'), ('safetyissue', 'New safety issues - notifies staff only'), ('buddyrequest', 'New buddy request - notifies all users'), ('buddyrequestmessage', 'New buddy request reply - notifies request creator and users who have replied')], help_text="Displays a the number of new notifications for the user. For example, if the user has two unread news notifications then the number '2' would appear for the news icon on the landing page.", max_length=25, null=True),
        ),
        migrations.CreateModel(
            name='BuddyRequestMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('content', models.TextField()),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('buddy_request', models.ForeignKey(help_text='The request that this message relates to.', on_delete=django.db.models.deletion.CASCADE, related_name='replies', to='NEMO.BuddyRequest')),
            ],
            options={
                'ordering': ['creation_date'],
            },
        ),
        migrations.CreateModel(
            name='EmailLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.IntegerField(choices=[(0, 'General'), (1, 'System'), (2, 'Direct Contact'), (3, 'Broadcast Email'), (4, 'Timed Services'), (5, 'Feedback'), (6, 'Abuse'), (7, 'Safety'), (8, 'Tasks')], default=0)),
                ('when', models.DateTimeField(auto_now_add=True)),
                ('sender', models.EmailField(max_length=254)),
                ('to', models.TextField()),
                ('subject', models.CharField(max_length=254)),
                ('content', models.TextField()),
                ('ok', models.BooleanField(default=True)),
                ('attachments', models.TextField(null=True)),
            ],
            options={
                'ordering': ['-when'],
            },
        ),
        migrations.RunPython(new_version_news),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-15 15:06
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0004_version_1_3_0'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='taskstatus',
            name='notify_secondary_tool_owner',
        ),
        migrations.RemoveField(
            model_name='tool',
            name='secondary_owner',
        ),
        migrations.AddField(
            model_name='taskstatus',
            name='notify_backup_tool_owners',
            field=models.BooleanField(default=False, help_text='Notify the backup tool owners when a task transitions to this status'),
        ),
        migrations.AddField(
            model_name='tool',
            name='backup_owners',
            field=models.ManyToManyField(blank=True, help_text='Alternate staff members who are responsible for administration of this tool when the primary owner is unavailable.', related_name='backup_for_tools', to=settings.AUTH_USER_MODEL),
        ),
    ]



import re
from logging import getLogger

from django.http import HttpResponseForbidden
from django.contrib.auth.middleware import RemoteUserMiddleware
from django.conf import settings
from django.urls import reverse, NoReverseMatch
from django.utils.deprecation import MiddlewareMixin

from NEMO.exceptions import InactiveUserError
from NEMO.models import User

middleware_logger = getLogger(__name__)


class RemoteUserAuthenticationMiddleware(RemoteUserMiddleware):

	def __init__(self, get_response=None):
		self.api_url = None
		try:
			self.api_url = reverse('api-root')
		except NoReverseMatch:
			pass
		super().__init__(get_response)

	def process_request(self, request):
		# REST API has its own authentication
		if request.path and self.api_url and request.path.startswith(self.api_url):
			return

		try:
			try:
				header_value = request.META[self.header]
			except KeyError:
				# If header is not present, log a warning and continue with processing of base class. (no authentication is happening)
				middleware_logger.debug(f"Header: {self.header} not present or invalid")
			super().process_request(request)
		except (User.DoesNotExist, InactiveUserError):
			from NEMO.views.authentication import all_auth_backends_are_pre_auth
			# Only raise error if all we have are pre_authentication backends and they failed
			if all_auth_backends_are_pre_auth():
				from NEMO.views.authentication import authorization_failed
				return authorization_failed(request)


class HTTPHeaderAuthenticationMiddleware(RemoteUserAuthenticationMiddleware):
	header = "HTTP_" + getattr(settings, "AUTHENTICATION_HEADER", "AUTHORIZATION")


class SessionTimeout:
	def __init__(self, get_response):
		self.get_response = get_response

	def __call__(self, request):
		return self.get_response(request)

	def process_view(self, request, view_function, view_ordered_arguments, view_named_arguments):
		"""
		This custom middleware exists to provide a smooth experience for public facing NEMO users.
		It processes all requests and redirects a user to the login page when their session times out.
		"""

		# If the user does not yet have a session then take no action.
		if not request.session:
			return None

		# If the user's session has timed out and this is an AJAX request then reply
		# with 403 forbidden because we don't want the AJAX request to produce the login page.
		# The base.html template has a global AJAX error callback to redirect to the logout
		# page when HTTP 403 is received.
		#
		# If the request is normal (instead of AJAX) and the user's session has expired
		# then the @login_required decorator will redirect them to the login page.
		if not request.user.is_authenticated:
			return HttpResponseForbidden() if request.is_ajax() else None

		# If the view is regularly polled by the webpage to update information then expiry refresh should be disabled.
		refresh_disabled = getattr(view_function, "disable_session_expiry_refresh", False)
		if not refresh_disabled:
			request.session.modified = True


class DeviceDetectionMiddleware:
	def __init__(self, get_response):
		self.get_response = get_response
		self.mobile = re.compile("Mobile|Tablet|Android")

	def __call__(self, request):
		request.device = "desktop"

		if "HTTP_USER_AGENT" in request.META:
			user_agent = request.META["HTTP_USER_AGENT"]
			if self.mobile.search(user_agent):
				request.device = "mobile"

		return self.get_response(request)


class ImpersonateMiddleware(MiddlewareMixin):
	def process_request(self, request):
		if request.user.is_superuser and 'impersonate_id' in request.session:
			request.user = User.objects.get(pk=request.session['impersonate_id'])



from datetime import datetime, timedelta
from decimal import Decimal
from typing import List, Optional

from django import forms
from django.db.models import QuerySet

from NEMO.models import (
	AreaAccessRecord,
	ConsumableWithdraw,
	Project,
	Reservation,
	StaffCharge,
	TrainingSession,
	UsageEvent,
	User,
)
from NEMO.utilities import localize


class BillingFilterForm(forms.Form):
	start = forms.DateField(required=True)
	end = forms.DateField(required=True)
	username = forms.CharField(required=False)
	account_name = forms.CharField(required=False)
	account_id = forms.IntegerField(required=False)
	application_name = forms.CharField(required=False)
	project_name = forms.CharField(required=False)
	project_id = forms.IntegerField(required=False)

	def get_start_date(self):
		return localize(datetime.combine(self.cleaned_data["start"], datetime.min.time()))

	def get_end_date(self):
		return localize(datetime.combine(self.cleaned_data["end"], datetime.max.time()))

	def get_project_id(self):
		return self.cleaned_data["project_id"]

	def get_project_name(self):
		return self.cleaned_data["project_name"]

	def get_account_id(self):
		return self.cleaned_data["account_id"]

	def get_account_name(self):
		return self.cleaned_data["account_name"]

	def get_username(self):
		return self.cleaned_data["username"]

	def get_application_name(self):
		return self.cleaned_data["application_name"]


class BillableItem(object):
	def __init__(self, item_type: str, project: Project, user: User):
		self.type: Optional[str] = item_type
		self.name: Optional[str] = None
		self.details: Optional[str] = ""
		if project:
			self.account: Optional[str] = project.account.name
			self.account_id: Optional[int] = project.account.id
			self.project: Optional[str] = project.name
			self.project_id: Optional[int] = project.id
			self.application: Optional[str] = project.application_identifier
		if user:
			self.user: Optional[str] = str(user)
			self.username: Optional[str] = user.username
			self.user_id: Optional[int] = user.id
		self.start: Optional[datetime] = None
		self.end: Optional[datetime] = None
		self.quantity: Optional[Decimal] = None


def get_usage_events_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = UsageEvent.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(end__gte=start, end__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(user__username=billing_form.get_username())
	return billable_items_usage_events(queryset)


def get_area_access_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = AreaAccessRecord.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(end__gte=start, end__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(customer__username=billing_form.get_username())
	return billable_items_area_access_records(queryset)


def get_missed_reservations_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = Reservation.objects.filter(missed=True)
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(end__gte=start, end__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(user__username=billing_form.get_username())
	return billable_items_missed_reservations(queryset)


def get_staff_charges_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = StaffCharge.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(end__gte=start, end__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(customer__username=billing_form.get_username())
	return billable_items_staff_charges(queryset)


def get_consumables_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = ConsumableWithdraw.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(date__gte=start, date__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(customer__username=billing_form.get_username())
	return billable_items_consumable_withdrawals(queryset)


def get_training_sessions_for_billing(billing_form: BillingFilterForm) -> List[BillableItem]:
	queryset = TrainingSession.objects.filter()
	start, end = billing_form.get_start_date(), billing_form.get_end_date()
	queryset = queryset.filter(date__gte=start, date__lte=end)
	if billing_form.get_account_id():
		queryset = queryset.filter(project__account_id=billing_form.get_account_id())
	if billing_form.get_account_name():
		queryset = queryset.filter(project__account__name=billing_form.get_account_name())
	if billing_form.get_project_id():
		queryset = queryset.filter(project__id=billing_form.get_project_id())
	if billing_form.get_project_name():
		queryset = queryset.filter(project__name=billing_form.get_project_name())
	if billing_form.get_application_name():
		queryset = queryset.filter(project__application_identifier=billing_form.get_application_name())
	if billing_form.get_username():
		queryset = queryset.filter(trainee__username=billing_form.get_username())
	return billable_items_training_sessions(queryset)


def billable_items_usage_events(usage_events: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	usage_event: UsageEvent
	for usage_event in usage_events:
		item = BillableItem("tool_usage", usage_event.project, usage_event.user)
		item.name = usage_event.tool.name
		item.details = (
			f"Work performed by {usage_event.operator} on user's behalf"
			if usage_event.operator != usage_event.user
			else ""
		)
		item.start = usage_event.start
		item.end = usage_event.end
		item.quantity = get_minutes_between_dates(usage_event.start, usage_event.end)
		billable_items.append(item)
	return billable_items


def billable_items_area_access_records(area_access_records: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	area_access_record: AreaAccessRecord
	for area_access_record in area_access_records:
		item = BillableItem("area_access", area_access_record.project, area_access_record.customer)
		item.name = area_access_record.area.name
		item.details = (
			f"Area accessed by {area_access_record.staff_charge.staff_member} on user's behalf"
			if area_access_record.staff_charge
			else ""
		)
		item.start = area_access_record.start
		item.end = area_access_record.end
		item.quantity = get_minutes_between_dates(area_access_record.start, area_access_record.end)
		billable_items.append(item)
	return billable_items


def billable_items_consumable_withdrawals(withdrawals: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	consumable_withdrawal: ConsumableWithdraw
	for consumable_withdrawal in withdrawals:
		item = BillableItem("consumable", consumable_withdrawal.project, consumable_withdrawal.customer)
		item.name = consumable_withdrawal.consumable.name
		item.start = consumable_withdrawal.date
		item.end = consumable_withdrawal.date
		item.quantity = consumable_withdrawal.quantity
		billable_items.append(item)
	return billable_items


def billable_items_missed_reservations(missed_reservations: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	missed_reservation: Reservation
	for missed_reservation in missed_reservations:
		item = BillableItem("missed_reservation", missed_reservation.project, missed_reservation.user)
		item.name = missed_reservation.reservation_item.name
		item.start = missed_reservation.start
		item.end = missed_reservation.end
		item.quantity = 1
		billable_items.append(item)
	return billable_items


def billable_items_staff_charges(staff_charges: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	staff_charge: StaffCharge
	for staff_charge in staff_charges:
		item = BillableItem("staff_charge", staff_charge.project, staff_charge.customer)
		item.details = staff_charge.note
		item.name = f"Work performed by {staff_charge.staff_member}"
		item.start = staff_charge.start
		item.end = staff_charge.end
		item.quantity = get_minutes_between_dates(staff_charge.start, staff_charge.end)
		billable_items.append(item)
	return billable_items


def billable_items_training_sessions(training_sessions: QuerySet) -> List[BillableItem]:
	billable_items: List[BillableItem] = []
	training_session: TrainingSession
	for training_session in training_sessions:
		item = BillableItem("training_session", training_session.project, training_session.trainee)
		item.name = training_session.tool.name
		item.details = f"{training_session.get_type_display()} training provided by {training_session.trainer}"
		item.start = training_session.date
		item.end = training_session.date
		item.quantity = training_session.duration
		billable_items.append(item)
	return billable_items


def get_minutes_between_dates(start, end, round_digits=2) -> Decimal:
	diff: timedelta = end - start
	return round(Decimal(diff.total_seconds()) / Decimal(60), round_digits)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-08-09 14:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0006_version_1_8_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='tool',
            name='policy_off_between_times',
            field=models.BooleanField(default=False, help_text='Check this box to disable the following policy rules every day between the given times'),
        ),
        migrations.AddField(
            model_name='tool',
            name='policy_off_end_time',
            field=models.TimeField(blank=True, help_text='The end time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='policy_off_start_time',
            field=models.TimeField(blank=True, help_text='The start time when policy rules should NOT be enforced', null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='policy_off_weekend',
            field=models.BooleanField(default=False, help_text='Whether or not policy rules should be enfored on weekends'),
        ),
    ]



from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.template import Context, Template
from django.urls import reverse
from django.views.decorators.http import require_GET, require_http_methods

from NEMO.decorators import staff_member_required
from NEMO.forms import SafetyIssueCreationForm, SafetyIssueUpdateForm
from NEMO.models import SafetyIssue
from NEMO.utilities import EmailCategory, send_mail
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.notifications import create_safety_notification, delete_notification, get_notifications


@login_required
@require_http_methods(["GET", "POST"])
def safety(request):
	dictionary = {}
	if request.method == "POST":
		form = SafetyIssueCreationForm(request.user, data=request.POST)
		if form.is_valid():
			issue = form.save()
			send_safety_email_notification(request, issue)
			dictionary["title"] = "Concern received"
			dictionary["heading"] = "Your safety concern was sent to the staff and will be addressed promptly"
			create_safety_notification(issue)
			return render(request, "acknowledgement.html", dictionary)
	tickets = SafetyIssue.objects.filter(resolved=False).order_by("-creation_time")
	if not request.user.is_staff:
		tickets = tickets.filter(visible=True)
	dictionary["tickets"] = tickets
	dictionary["safety_introduction"] = get_media_file_contents("safety_introduction.html")
	dictionary["notifications"] = get_notifications(request.user, SafetyIssue)
	return render(request, "safety/safety.html", dictionary)


def send_safety_email_notification(request, issue):
	recipient = get_customization("safety_email_address")
	message = get_media_file_contents("safety_issue_email.html")
	if recipient and message:
		subject = "Safety issue"
		dictionary = {"issue": issue, "issue_absolute_url": request.build_absolute_uri(issue.get_absolute_url())}
		rendered_message = Template(message).render(Context(dictionary))
		from_email = issue.reporter.email if issue.reporter else recipient
		send_mail(
			subject=subject,
			content=rendered_message,
			from_email=from_email,
			to=[recipient],
			email_category=EmailCategory.SAFETY,
		)


@login_required
@require_GET
def resolved_safety_issues(request):
	tickets = SafetyIssue.objects.filter(resolved=True)
	if not request.user.is_staff:
		tickets = tickets.filter(visible=True)
	return render(request, "safety/resolved_issues.html", {"tickets": tickets})


@staff_member_required
@require_http_methods(["GET", "POST"])
def update_safety_issue(request, ticket_id):
	if request.method == "POST":
		ticket = get_object_or_404(SafetyIssue, id=ticket_id)
		form = SafetyIssueUpdateForm(request.user, data=request.POST, instance=ticket)
		if form.is_valid():
			issue = form.save()
			if issue.resolved:
				delete_notification(SafetyIssue, issue.id)
			return HttpResponseRedirect(reverse("safety"))
	dictionary = {"ticket": get_object_or_404(SafetyIssue, id=ticket_id)}
	return render(request, "safety/update_issue.html", dictionary)



from django.core.files.storage import get_storage_class
from django.core.validators import validate_email
from django.http import HttpResponseBadRequest
from django.shortcuts import redirect, render
from django.views.decorators.http import require_GET, require_POST

from NEMO import init_admin_site
from NEMO.decorators import administrator_required
from NEMO.exceptions import InvalidCustomizationException
from NEMO.models import Customization


def get_media_file_contents(file_name):
	""" Get the contents of a media file if it exists. Return a blank string if it does not exist. """
	storage = get_storage_class()()
	if not storage.exists(file_name):
		return ''
	f = storage.open(file_name)
	try:
		return f.read().decode().strip()
	except UnicodeDecodeError:
		f = storage.open(file_name)
		return f.read()


def store_media_file(content, file_name):
	""" Delete any existing media file with the same name and save the new content into file_name in the media directory. If content is blank then no new file is created. """
	storage = get_storage_class()()
	storage.delete(file_name)
	if content:
		storage.save(file_name, content)


# Dictionary of customizable keys with default value
customizable_key_values = {
	'feedback_email_address': '',
	'user_office_email_address': '',
	'safety_email_address': '',
	'abuse_email_address': '',
	'facility_name': 'NanoFab',
	'site_title': 'NEMO',
	'self_log_in': '',
	'self_log_out': '',
	'calendar_login_logout': '',
	'dashboard_display_not_qualified_areas': '',
	'dashboard_staff_status_first_day_of_week': '1',
	'dashboard_staff_status_staff_only': '',
	'dashboard_staff_status_weekdays_only': '',
	'dashboard_staff_status_date_format': 'D m/d',
	'calendar_view': 'agendaWeek',
	'calendar_first_day_of_week': '1',
	'calendar_day_column_format': 'dddd MM/DD/YYYY',
	'calendar_week_column_format': 'ddd M/DD',
	'calendar_month_column_format': 'ddd',
	'calendar_start_of_the_day': '07:00:00',
	'calendar_now_indicator': '',
	'calendar_display_not_qualified_areas': '',
	'calendar_all_tools': '',
	'calendar_all_areas': '',
	'calendar_all_areastools': '',
	'calendar_outage_recurrence_limit': '90',
	'project_selection_template': '{{ project.name }}',
	'allow_bypass_interlock_on_failure': '',
	'tool_interlock_failure_message': 'Communication with the interlock failed',
	'door_interlock_failure_message': 'Communication with the interlock failed',
	'buddy_requests_title': 'Buddy requests board',
	'buddy_board_description': '',
	'access_requests_title': 'Access requests',
	'access_requests_description': '',
	'access_requests_minimum_users': '2',
	'access_requests_display_max': '',
	'weekend_access_notification_emails': '',
	'weekend_access_notification_cutoff_hour': '',
	'weekend_access_notification_cutoff_day': '',
	'weekend_access_notification_last_sent': '',
}

customizable_content = [
	('access_request_notification_email', '.html'),
	('authorization_failed', '.html'),
	('cancellation_email', '.html'),
	('counter_threshold_reached_email', '.html'),
	('facility_rules_tutorial', '.html'),
	('facility_rules_tutorial_email', '.html'),
	('feedback_email', '.html'),
	('generic_email', '.html'),
	('login_banner', '.html'),
	('missed_reservation_email', '.html'),
	('new_task_email', '.html'),
	('out_of_time_reservation_email', '.html'),
	('reorder_supplies_reminder_email', '.html'),
	('reservation_cancelled_user_email', '.html'),
	('reservation_created_user_email', '.html'),
	('reservation_ending_reminder_email', '.html'),
	('reservation_reminder_email', '.html'),
	('reservation_warning_email', '.html'),
	('safety_introduction', '.html'),
	('safety_issue_email', '.html'),
	('staff_charge_reminder_email', '.html'),
	('task_status_notification', '.html'),
	('unauthorized_tool_access_email', '.html'),
	('usage_reminder_email', '.html'),
	('weekend_access_email', '.html'),
	('weekend_no_access_email', '.html'),
	('rates', '.json'),
	('jumbotron_watermark', '.png'),
]


def get_customization(name, raise_exception=True):
	if name not in customizable_key_values.keys():
		raise InvalidCustomizationException(name)
	default_value = customizable_key_values[name]
	try:
		return Customization.objects.get(name=name).value
	except Customization.DoesNotExist:
		# return default value
		return default_value
	except Exception:
		if raise_exception:
			raise
		else:
			return default_value


def set_customization(name, value):
	if name not in customizable_key_values.keys():
		raise InvalidCustomizationException(name, value)
	if value:
		if name in ['feedback_email_address', 'user_office_email_address', 'safety_email_address', 'abuse_email_address']:
			validate_email(value)
		Customization.objects.update_or_create(name=name, defaults={'value': value})
	else:
		try:
			Customization.objects.get(name=name).delete()
		except Customization.DoesNotExist:
			pass


@administrator_required
@require_GET
def customization(request):
	dictionary = {name: get_media_file_contents(name + extension) for name, extension in customizable_content}
	dictionary.update({name: get_customization(name) for name in customizable_key_values.keys()})
	return render(request, 'customizations/customizations.html', dictionary)


@administrator_required
@require_POST
def customize(request, element):
	item = None
	for name, extension in customizable_content:
		if name == element:
			item = (name, extension)
			break
	if item:
		store_media_file(request.FILES.get(element, ''), item[0] + item[1])
		if item[0] == 'rates':
			from NEMO.rates import rate_class
			rate_class.load_rates(force_reload=True)
	elif element == 'email_addresses':
		set_customization('feedback_email_address', request.POST.get('feedback_email_address', ''))
		set_customization('safety_email_address', request.POST.get('safety_email_address', ''))
		set_customization('abuse_email_address', request.POST.get('abuse_email_address', ''))
		set_customization('user_office_email_address', request.POST.get('user_office_email_address', ''))
	elif element == 'application_settings':
		set_customization('self_log_in', request.POST.get('self_log_in', ''))
		set_customization('self_log_out', request.POST.get('self_log_out', ''))
		set_customization('calendar_login_logout', request.POST.get('calendar_login_logout', ''))
		set_customization('facility_name', request.POST.get('facility_name', ''))
		set_customization('site_title', request.POST.get('site_title', ''))
		set_customization('project_selection_template', request.POST.get('project_selection_template', ''))
		init_admin_site()
	elif element == 'calendar_settings':
		set_customization('calendar_view', request.POST.get('calendar_view', ''))
		set_customization('calendar_first_day_of_week', request.POST.get('calendar_first_day_of_week', ''))
		set_customization('calendar_start_of_the_day', request.POST.get('calendar_start_of_the_day', ''))
		set_customization('calendar_now_indicator', request.POST.get('calendar_now_indicator', ''))
		set_customization('calendar_day_column_format', request.POST.get('calendar_day_column_format', ''))
		set_customization('calendar_week_column_format', request.POST.get('calendar_week_column_format', ''))
		set_customization('calendar_month_column_format', request.POST.get('calendar_month_column_format', ''))
		set_customization('calendar_all_tools', request.POST.get('calendar_all_tools', ''))
		set_customization('calendar_all_areas', request.POST.get('calendar_all_areas', ''))
		set_customization('calendar_all_areastools', request.POST.get('calendar_all_areastools', ''))
		set_customization('calendar_outage_recurrence_limit', request.POST.get('calendar_outage_recurrence_limit', '90'))
	elif element == 'dashboard_settings':
		set_customization('dashboard_display_not_qualified_areas', request.POST.get('dashboard_display_not_qualified_areas', ''))
		set_customization('dashboard_staff_status_first_day_of_week', request.POST.get('dashboard_staff_status_first_day_of_week', ''))
		set_customization('dashboard_staff_status_staff_only', request.POST.get('dashboard_staff_status_staff_only', ''))
		set_customization('dashboard_staff_status_weekdays_only', request.POST.get('dashboard_staff_status_weekdays_only', ''))
		set_customization('dashboard_staff_status_date_format', request.POST.get('dashboard_staff_status_date_format', ''))
	elif element == 'interlock_settings':
		set_customization('allow_bypass_interlock_on_failure', request.POST.get('allow_bypass_interlock_on_failure', ''))
		set_customization('tool_interlock_failure_message', request.POST.get('tool_interlock_failure_message', ''))
		set_customization('door_interlock_failure_message', request.POST.get('door_interlock_failure_message', ''))
	elif element == 'requests_settings':
		set_customization('buddy_requests_title', request.POST.get('buddy_requests_title', ''))
		set_customization('buddy_board_description', request.POST.get('buddy_board_description', ''))
		set_customization('access_requests_title', request.POST.get('access_requests_title', ''))
		set_customization('access_requests_description', request.POST.get('access_requests_description', ''))
		set_customization('access_requests_minimum_users', request.POST.get('access_requests_minimum_users', ''))
		set_customization('access_requests_display_max', request.POST.get('access_requests_display_max', ''))
		set_customization('weekend_access_notification_emails', request.POST.get('weekend_access_notification_emails', ''))
		set_customization('weekend_access_notification_cutoff_hour', request.POST.get('weekend_access_notification_cutoff_hour', ''))
		set_customization('weekend_access_notification_cutoff_day', request.POST.get('weekend_access_notification_cutoff_day', ''))
	else:
		return HttpResponseBadRequest('Invalid customization')
	return redirect('customization')



# Generated by Django 2.2.26 on 2022-02-07 19:08

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

from NEMO.migrations_utils import create_news_for_version


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0035_version_3_14_0'),
    ]

    def new_version_news(apps, schema_editor):
        create_news_for_version(apps, "3.15.0")

    def migrate_exceptions_to_closures(apps, schema_editor):
        Closure = apps.get_model("NEMO", "Closure")
        ClosureTime = apps.get_model("NEMO", "ClosureTime")
        try:
            for closure in Closure.objects.all():
                ClosureTime.objects.create(closure=closure, start_time=closure.start_time, end_time=closure.end_time)
        except:
            pass

    operations = [
        migrations.RunPython(new_version_news),
        migrations.CreateModel(
            name='StaffAbsenceType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of this absence type.', max_length=255)),
                ('description', models.CharField(help_text='The description for this absence type.', max_length=255)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='StaffAvailability',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, help_text='The category for this staff member.', max_length=100, null=True)),
                ('start_time', models.TimeField(blank=True, help_text='The usual start time for this staff member.', null=True)),
                ('end_time', models.TimeField(blank=True, help_text='The usual end time for this staff member.', null=True)),
                ('monday', models.BooleanField(default=True, help_text='Check this box if the staff member usually works on Mondays.')),
                ('tuesday', models.BooleanField(default=True, help_text='Check this box if the staff member usually works on Tuesdays.')),
                ('wednesday', models.BooleanField(default=True, help_text='Check this box if the staff member usually works on Wednesdays.')),
                ('thursday', models.BooleanField(default=True, help_text='Check this box if the staff member usually works on Thursdays.')),
                ('friday', models.BooleanField(default=True, help_text='Check this box if the staff member usually works on Fridays.')),
                ('saturday', models.BooleanField(default=False, help_text='Check this box if the staff member usually works on Saturdays.')),
                ('sunday', models.BooleanField(default=False, help_text='Check this box if the staff member usually works on Sundays.')),
                ('staff_member', models.ForeignKey(help_text='The staff member to display on the staff status page.', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'Staff availability',
                'ordering': ['staff_member__first_name'],
            },
        ),
        migrations.CreateModel(
            name='StaffAbsence',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_time', models.DateTimeField(auto_now_add=True, help_text="The date and time when the absence was created.")),
                ('start_date', models.DateField(help_text='The start date of the absence.')),
                ('end_date', models.DateField(help_text='The end date of the absence.')),
                ('full_day', models.BooleanField(default=True, help_text='Uncheck this box when the absence is only for part of the day.')),
                ('description', models.TextField(blank=True, help_text='The absence description. This will be visible to anyone when the absence is not all day.', null=True)),
                ('absence_type', models.ForeignKey(help_text='The absence type. This will only be visible to facility managers.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.StaffAbsenceType')),
                ('staff_member', models.ForeignKey(help_text='The staff member who will be absent.', on_delete=django.db.models.deletion.CASCADE, to='NEMO.StaffAvailability')),
            ],
            options={
                'ordering': ['-creation_time']
            }
        ),
        migrations.RenameModel(old_name='PhysicalAccessException', new_name='Closure'),
        migrations.AlterField(
            model_name='closure',
            name='name',
            field=models.CharField(help_text='The name of this closure, that will be displayed as the policy problem and alert (if applicable)', max_length=255),
        ),
        migrations.CreateModel(
            name='ClosureTime',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.DateTimeField(help_text='The start date and time of the closure')),
                ('end_time', models.DateTimeField(help_text='The end date and time of the closure')),
                ('closure', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='NEMO.Closure')),
            ],
            options={
                'ordering': ['-start_time'],
            },
        ),
        migrations.RunPython(migrate_exceptions_to_closures),
        migrations.AlterModelOptions(
            name='closure',
            options={'ordering': ['name']},
        ),
        migrations.RemoveField(
            model_name='closure',
            name='end_time',
        ),
        migrations.RemoveField(
            model_name='closure',
            name='start_time',
        ),
        migrations.AddField(
            model_name='closure',
            name='alert_days_before',
            field=models.PositiveIntegerField(blank=True, help_text='Enter the number of days before the closure when an alert should automatically be created. Leave blank for no alert.', null=True),
        ),
        migrations.AddField(
            model_name='closure',
            name='alert_template',
            field=models.TextField(blank=True, help_text='The template to create the alert with. The following variables are provided (when applicable): <b>name</b>, <b>start_time</b>, <b>end_time</b>, <b>areas</b>.', null=True),
        ),
        migrations.AddField(
            model_name='closure',
            name='staff_absent',
            field=models.BooleanField(default=True, help_text='Check this box and all staff members will be marked absent during this closure in staff status.'),
        ),
        migrations.AlterField(
            model_name='closure',
            name='name',
            field=models.CharField(help_text='The name of this closure, that will be displayed as the policy problem and alert (if applicable).', max_length=255),
        ),
        migrations.AlterField(
            model_name='closure',
            name='physical_access_levels',
            field=models.ManyToManyField(blank=True, help_text='Select access levels this closure applies to.', to='NEMO.PhysicalAccessLevel'),
        ),
        migrations.AddField(
            model_name='closure',
            name='notify_managers_last_occurrence',
            field=models.BooleanField(default=True, help_text='Check this box to notify facility managers on the last occurrence of this closure.'),
        ),
    ]



from django.contrib import messages
from django.urls import reverse
from django.utils.safestring import mark_safe

from NEMO.models import User, Area, Tool


def lock_selected_interlocks(model_admin, request, queryset):
	for interlock in queryset:
		try:
			interlock.lock()
			messages.success(request, f"{interlock} has been successfully locked")
		except Exception as error:
			messages.error(request, f"{interlock} could not be locked due to the following error: {str(error)}")


def unlock_selected_interlocks(model_admin, request, queryset):
	for interlock in queryset:
		try:
			interlock.unlock()
			messages.success(request, f"{interlock} has been successfully unlocked")
		except Exception as error:
			messages.error(request, f"{interlock} could not be unlocked due to the following error: {str(error)}")


def synchronize_with_tool_usage(model_admin, request, queryset):
	for interlock in queryset:
		# Ignore interlocks with no tool assigned, and ignore interlocks connected to doors
		if not hasattr(interlock, "tool") or hasattr(interlock, "door"):
			continue
		if interlock.tool.in_use():
			interlock.unlock()
		else:
			interlock.lock()


def duplicate_tool_configuration(model_admin, request, queryset):
	for tool in queryset:
		original_name = tool.name
		new_name = "Copy of " + tool.name
		try:
			if Tool.objects.filter(name=new_name).exists():
				messages.error(
					request,
					mark_safe(
						f'There is already a copy of {original_name} as <a href="{reverse("admin:NEMO_tool_change", args=[tool.id])}">{new_name}</a>. Change the copy\'s name and try again'
					),
				)
				continue
			elif tool.is_child_tool():
				messages.warning(request, f"{original_name} is a child tool and cannot be duplicated")
				continue
			else:
				old_required_resources = tool.required_resource_set.all()
				old_nonrequired_resources = tool.nonrequired_resource_set.all()
				old_backup_users = tool.backup_owners.all()
				old_superusers = tool.superusers.all()
				old_qualified_users = User.objects.filter(qualifications__id=tool.pk).distinct()
				tool.pk = None
				tool.interlock = None
				tool.visible = False
				tool.operational = False
				tool.name = new_name
				tool.image = None
				tool.description = None
				tool.serial = None
				tool.save()
				tool.required_resource_set.set(old_required_resources)
				tool.nonrequired_resource_set.set(old_nonrequired_resources)
				tool.backup_owners.set(old_backup_users)
				tool.superusers.set(old_superusers)
				for user in old_qualified_users:
					user.qualifications.add(tool)
				messages.success(
					request,
					mark_safe(
						f'A duplicate of {original_name} has been made as <a href="{reverse("admin:NEMO_tool_change", args=[tool.id])}">{tool.name}</a>'
					),
				)
		except Exception as error:
			messages.error(
				request, f"{original_name} could not be duplicated because of the following error: {str(error)}"
			)


def rebuild_area_tree(model_admin, request, queryset):
	Area.objects.rebuild()



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.views.decorators.http import require_GET

from NEMO.models import PhysicalAccessLevel, User
from NEMO.views.customization import get_customization


@login_required
@require_GET
def user_requests(request, tab: str = None):
	active_tab = (
		tab or "access"
		if PhysicalAccessLevel.objects.filter(allow_user_request=True).exists()
		   and User.objects.filter(is_active=True, is_facility_manager=True).exists()
		else "buddy"
	)
	buddy_requests_title = get_customization("buddy_requests_title")
	access_requests_title = get_customization("access_requests_title")
	dictionary = {
		"tab": active_tab,
		"buddy_requests_title": buddy_requests_title,
		"access_requests_title": access_requests_title,
	}
	return render(request, "requests/user_requests.html", dictionary)



from collections import defaultdict
from datetime import date, datetime, timedelta
from typing import List, Optional, Union

from django.db.models import Q
from django.http import HttpResponse, HttpResponseBadRequest
from django.template import Context, Template
from django.utils import timezone

from NEMO.exceptions import (
	InactiveUserError,
	ItemNotAllowedForProjectException,
	MaximumCapacityReachedError,
	NoAccessiblePhysicalAccessUserError,
	NoActiveProjectsForUserError,
	NoPhysicalAccessUserError,
	NotAllowedToChargeProjectException,
	PhysicalAccessExpiredUserError,
	ProjectChargeException,
	ReservationRequiredUserError,
	ScheduledOutageInProgressError,
	UnavailableResourcesUserError,
)
from NEMO.models import (
	Area,
	AreaAccessRecord,
	ClosureTime,
	Consumable,
	PhysicalAccessLevel,
	Project,
	Reservation,
	ReservationItemType,
	ScheduledOutage,
	StaffCharge,
	Tool,
	User,
)
from NEMO.utilities import EmailCategory, distinct_qs_value_list, format_daterange, format_datetime, send_mail
from NEMO.views.customization import get_customization, get_media_file_contents


def check_policy_to_enable_tool(tool: Tool, operator: User, user: User, project: Project, staff_charge: bool):
	"""
	Check that the user is allowed to enable the tool. Enable the tool if the policy checks pass.
	"""
	facility_name = get_customization('facility_name')

	# The tool must be visible (or the parent if it's a child tool) to users.
	visible = tool.parent_tool.visible if tool.is_child_tool() else tool.visible
	if not visible:
		return HttpResponseBadRequest("This tool is currently hidden from users.")

	# The tool must be operational.
	# If the tool is non-operational then it may only be accessed by staff members or service personnel.
	if not tool.operational and not operator.is_staff and not operator.is_service_personnel:
		return HttpResponseBadRequest("This tool is currently non-operational.")

	# The tool must not be in use.
	current_usage_event = tool.get_current_usage_event()
	if current_usage_event:
		return HttpResponseBadRequest("The tool is currently being used by " + str(current_usage_event.user) + ".")

	# The user must be qualified to use the tool itself, or the parent tool in case of alternate tool.
	tool_to_check_qualifications = tool.parent_tool if tool.is_child_tool() else tool
	if tool_to_check_qualifications not in operator.qualifications.all() and not operator.is_staff:
		return HttpResponseBadRequest("You are not qualified to use this tool.")

	# Only staff members can operate a tool on behalf of another user.
	if (user and operator.pk != user.pk) and not operator.is_staff:
		return HttpResponseBadRequest("You must be a staff member to use a tool on another user's behalf.")

	# All required resources must be available to operate a tool except for staff or service personnel.
	if tool.required_resource_set.filter(available=False).exists() and not operator.is_staff and not operator.is_service_personnel:
		return HttpResponseBadRequest("A resource that is required to operate this tool is unavailable.")

	# The tool operator may not activate tools in a particular area unless they are logged in to the area.
	# Staff are exempt from this rule.
	if tool.requires_area_access and AreaAccessRecord.objects.filter(area=tool.requires_area_access, customer=operator, staff_charge=None, end=None).count() == 0 and not operator.is_staff:
		abuse_email_address = get_customization('abuse_email_address')
		message = get_media_file_contents('unauthorized_tool_access_email.html')
		if abuse_email_address and message:
			dictionary = {
				'operator': operator,
				'tool': tool,
				'type': 'access'
			}
			rendered_message = Template(message).render(Context(dictionary))
			send_mail(subject="Area access requirement", content=rendered_message, from_email=abuse_email_address, to=[abuse_email_address], email_category=EmailCategory.ABUSE)
		return HttpResponseBadRequest("You must be logged in to the {} to operate this tool.".format(tool.requires_area_access.name))

	# The tool operator may not activate tools in a particular area unless they are still within that area reservation window
	# Staff and service personnel are exempt from this rule.
	if not operator.is_staff and not operator.is_service_personnel and tool.requires_area_reservation():
		if not tool.requires_area_access.get_current_reservation_for_user(operator):
			abuse_email_address = get_customization('abuse_email_address')
			message = get_media_file_contents('unauthorized_tool_access_email.html')
			if abuse_email_address and message:
				dictionary = {
					'operator': operator,
					'tool': tool,
					'type': 'reservation',
				}
				rendered_message = Template(message).render(Context(dictionary))
				send_mail(subject="Area reservation requirement", content=rendered_message, from_email=abuse_email_address, to=[abuse_email_address], email_category=EmailCategory.ABUSE)
			return HttpResponseBadRequest("You must have a current reservation for the {} to operate this tool.".format(tool.requires_area_access.name))

	# Staff may only charge staff time for one user at a time.
	if staff_charge and operator.charging_staff_time():
		return HttpResponseBadRequest('You are already charging staff time. You must end the current staff charge before you being another.')

	# Staff may not bill staff time to themselves.
	if staff_charge and operator == user:
		return HttpResponseBadRequest('You cannot charge staff time to yourself.')

	# Check if we are allowed to bill to project
	try:
		check_billing_to_project(project, user, tool)
	except ProjectChargeException as e:
		return HttpResponseBadRequest(e.msg)

	# The tool operator must not have a lock on usage
	if operator.training_required:
		return HttpResponseBadRequest(f"You are blocked from using all tools in the {facility_name}. Please complete the {facility_name} rules tutorial in order to use tools.")

	# Users may only use a tool when delayed logoff is not in effect. Staff and service personnel are exempt from this rule.
	if tool.delayed_logoff_in_progress() and not operator.is_staff and not operator.is_service_personnel:
		return HttpResponseBadRequest("Delayed tool logoff is in effect. You must wait for the delayed logoff to expire before you can use the tool.")

	# Users may not enable a tool during a scheduled outage. Staff and service personnel are exempt from this rule.
	if tool.scheduled_outage_in_progress() and not operator.is_staff and not operator.is_service_personnel:
		return HttpResponseBadRequest("A scheduled outage is in effect. You must wait for the outage to end before you can use the tool.")

	return HttpResponse()


def check_policy_to_disable_tool(tool, operator, downtime):
	""" Check that the user is allowed to disable the tool. """
	current_usage_event = tool.get_current_usage_event()
	if current_usage_event.operator != operator and current_usage_event.user != operator and not operator.is_staff:
		return HttpResponseBadRequest('You may not disable a tool while another user is using it unless you are a staff member.')
	if downtime < timedelta():
		return HttpResponseBadRequest('Downtime cannot be negative.')
	if downtime > timedelta(minutes=120):
		return HttpResponseBadRequest('Post-usage tool downtime may not exceed 120 minutes.')
	if tool.delayed_logoff_in_progress() and downtime > timedelta():
		return HttpResponseBadRequest('The tool is already in a delayed-logoff state. You may not issue additional delayed logoffs until the existing one expires.')
	if not tool.allow_delayed_logoff and downtime > timedelta():
		return HttpResponseBadRequest('Delayed logoff is not allowed for this tool.')
	return HttpResponse()


def check_policy_to_save_reservation(cancelled_reservation: Optional[Reservation], new_reservation: Reservation, user_creating_reservation: User, explicit_policy_override: bool):
	"""
		Check the reservation creation policy and return a list of policy problems if any.
	"""
	user = new_reservation.user

	facility_name = get_customization('facility_name')

	# The function will check all policies. Policy problems are placed in the policy_problems list. overridable is True if the policy problems can be overridden by a staff member.
	policy_problems = []
	overridable = False

	item_type = new_reservation.reservation_item_type

	# Reservations may not have a start time that is earlier than the end time.
	if new_reservation.start >= new_reservation.end:
		policy_problems.append("Reservation start time (" + format_datetime(new_reservation.start) + ") must be before the end time (" + format_datetime(new_reservation.end) + ").")

	check_coincident_item_reservation_policy(cancelled_reservation, new_reservation, user_creating_reservation, policy_problems)

	# Reservations that have been cancelled may not be changed.
	if new_reservation.cancelled:
		policy_problems.append("This reservation has already been cancelled by " + str(new_reservation.cancelled_by) + " at " + format_datetime(new_reservation.cancellation_time) + ".")

	# The user must belong to at least one active project to make a reservation.
	if user.active_project_count() < 1:
		if user == user_creating_reservation:
			policy_problems.append("You do not belong to any active projects. Thus, you may not create any reservations.")
		else:
			policy_problems.append(str(user) + " does not belong to any active projects and cannot have reservations.")

	# Check if we are allowed to bill to project
	try:
		check_billing_to_project(new_reservation.project, user, new_reservation.reservation_item)
	except ProjectChargeException as e:
		policy_problems.append(e.msg)

	# If the user is a staff member or there's an explicit policy override then the policy check is finished.
	if user.is_staff or explicit_policy_override:
		return policy_problems, overridable

	# If there are no blocking policy conflicts at this point, the rest of the policies can be overridden.
	if not policy_problems:
		overridable = True

	# Some tool reservations require a prior area reservation
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item_type == ReservationItemType.TOOL:
		if new_reservation.tool.requires_area_reservation():
			area: Area = new_reservation.tool.requires_area_access
			# Check that a reservation for the area has been made and contains the start time
			if not Reservation.objects.filter(missed=False, cancelled=False, shortened=False,
											  user=user, area=area,
											  start__lte=new_reservation.start,
											  end__gt=new_reservation.start).exists():
				if user == user_creating_reservation:
					policy_problems.append(f"This tool requires a {area} reservation. Please make a reservation in the {area} prior to reserving this tool.")
				else:
					policy_problems.append(f"This tool requires a {area} reservation. Please make sure to also create a reservation in the {area} or {str(user)} will not be able to enter the area.")

	# The user must complete training to create reservations.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if user.training_required:
		if user == user_creating_reservation:
			policy_problems.append(f"You are blocked from making reservations in the {facility_name}. Please complete the {facility_name} rules tutorial in order to create new reservations.")
		else:
			policy_problems.append(f"{str(user)} is blocked from making reservations in the {facility_name}. The user needs to complete the {facility_name} rules tutorial in order to create new reservations.")

	# Users may only change their own reservations.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if cancelled_reservation and user != user_creating_reservation:
		policy_problems.append("You may not change reservations that you do not own.")

	# The user may not create or move a reservation to have a start time that is earlier than the current time.
	# Unless it's an extension of an area reservation, in which case the start time is the same.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	extension_of_area_reservation = new_reservation.area and cancelled_reservation and cancelled_reservation.start == new_reservation.start
	if not extension_of_area_reservation and new_reservation.start < timezone.now():
		policy_problems.append("Reservation start time (" + format_datetime(new_reservation.start) + ") is earlier than the current time (" + format_datetime() + ").")

	# The user may not move or resize a reservation to have an end time that is earlier than the current time.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.end < timezone.now():
		policy_problems.append("Reservation end time (" + format_datetime(new_reservation.end) + ") is earlier than the current time (" + format_datetime() + ").")

	# The user must be qualified on the tool in question in order to create, move, or resize a reservation.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if new_reservation.tool and new_reservation.tool not in user.qualifications.all():
		if user == user_creating_reservation:
			policy_problems.append("You are not qualified to use this tool. Creating, moving, and resizing reservations is forbidden.")
		else:
			policy_problems.append(f"{str(user)} is not qualified to use this tool. Creating, moving, and resizing reservations is forbidden.")

	# The user must be authorized on the area in question at the start and end times of the reservation in order to create, move, or resize a reservation.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item_type == ReservationItemType.AREA:
		user_access_levels = user.accessible_access_levels_for_area(new_reservation.area)
		if not any([access_level.accessible_at(new_reservation.start) for access_level in user_access_levels]) or not any([access_level.accessible_at(new_reservation.end) for access_level in user_access_levels]):
			# it could be inaccessible because of an ongoing closure at the start or end time
			first_closure_time: ClosureTime = next(iter([access_level.ongoing_closure_time(new_reservation.start) for access_level in user_access_levels]), None)
			if not first_closure_time:
				# if there is no closure at start, check at end time
				first_closure_time = next(iter([access_level.ongoing_closure_time(new_reservation.end) for access_level in user_access_levels]), None)
			if first_closure_time:
				details = f" due to the following closure: {first_closure_time.closure.name} ({format_daterange(first_closure_time.start_time, first_closure_time.end_time)})"
			# or simply due to scheduling
			else:
				details = f" (times allowed in this area are: {', '.join([access.get_schedule_display_with_times() for access in user_access_levels])})" if user_access_levels else ''
			if user == user_creating_reservation:
				policy_problems.append(f"You are not authorized to access this area at this time{details}. Creating, moving, and resizing reservations is forbidden.")
			else:
				policy_problems.append(f"{str(user)} is not authorized to access this area at this time{details}. Creating, moving, and resizing reservations is forbidden.")

	check_tool_reservation_requiring_area(policy_problems, user_creating_reservation, cancelled_reservation, new_reservation)

	# The reservation start time may not exceed the item's reservation horizon.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	item = new_reservation.reservation_item
	if item.reservation_horizon is not None:
		reservation_horizon = timedelta(days=item.reservation_horizon)
		if new_reservation.start > timezone.now() + reservation_horizon:
			policy_problems.append("You may not create reservations further than " + str(reservation_horizon.days) + f" days from now for this {item_type.value}.")

	# Check item policy rules
	item_policy_problems = []
	if should_enforce_policy(new_reservation):
		item_policy_problems = check_policy_rules_for_item(user_creating_reservation, new_reservation, cancelled_reservation)

	# Return the list of all policies that are not met.
	return policy_problems + item_policy_problems, overridable


def check_tool_reservation_requiring_area(policy_problems: List[str], user_creating_reservation: User, cancelled_reservation: Optional[Reservation], new_reservation: Optional[Reservation]):
	# When modifying an area reservation, check that all tools reservations starting during the cancelled one are still within the new one
	if cancelled_reservation and cancelled_reservation.reservation_item_type == ReservationItemType.AREA:
		tools_requiring_cancelled_reservation = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, tool__isnull=False, tool___requires_area_access=cancelled_reservation.area, user=cancelled_reservation.user)
		tools_requiring_cancelled_reservation = tools_requiring_cancelled_reservation.filter(start__gte=cancelled_reservation.start, start__lt=cancelled_reservation.end)
		if tools_requiring_cancelled_reservation.exists():
			if new_reservation:
				tools_requiring_new_reservation = Reservation.objects.filter(cancelled=False, missed=False, shortened=False, tool__isnull=False, tool___requires_area_access=new_reservation.area, user=new_reservation.user)
				tools_requiring_new_reservation = tools_requiring_new_reservation.filter(start__gte=new_reservation.start, start__lt=new_reservation.end)
			else:
				tools_requiring_new_reservation = Reservation.objects.none()
			difference: List[Reservation] = [item for item in tools_requiring_cancelled_reservation if item not in tools_requiring_new_reservation]
			# As long as the new reservation includes the same tool reservations, we are good
			if difference:
				user = cancelled_reservation.user
				area = new_reservation.area if new_reservation else cancelled_reservation.area
				difference.sort(key=lambda x: x.start)
				if user == user_creating_reservation:
					policy_problems.append(f"You have a reservation for the {difference[0].tool} at {format_datetime(difference[0].start)} that requires a {area} reservation. Cancel or reschedule the tool reservation first and try again.")
				else:
					policy_problems.append(f"{str(user)} has a reservation for the {difference[0].tool} at {format_datetime(difference[0].start)} that requires a {area} reservation. Cancel or reschedule the tool reservation first and try again.")


def check_coincident_item_reservation_policy(cancelled_reservation: Optional[Reservation], new_reservation: Reservation, user_creating_reservation: User, policy_problems: List):
	user = new_reservation.user

	# For tools the user may not create, move, or resize a reservation to coincide with another user's reservation.
	# For areas, it cannot coincide with another reservation for the same user, or with a number of other users greater than the area capacity
	coincident_events = Reservation.objects.filter(cancelled=False, missed=False, shortened=False)
	# Exclude the reservation we're cancelling in order to create a new one:
	if cancelled_reservation and cancelled_reservation.id:
		coincident_events = coincident_events.exclude(id=cancelled_reservation.id)
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=new_reservation.start, end__lte=new_reservation.start)
	coincident_events = coincident_events.exclude(start__gte=new_reservation.end, end__gt=new_reservation.end)
	if new_reservation.reservation_item_type == ReservationItemType.TOOL and coincident_events.filter(**new_reservation.reservation_item_filter).count() > 0:
		policy_problems.append("Your reservation coincides with another reservation that already exists. Please choose a different time.")
	if new_reservation.reservation_item_type == ReservationItemType.AREA:
		if coincident_events.filter(**new_reservation.reservation_item_filter).filter(user=user).count() > 0:
			if user == user_creating_reservation:
				policy_problems.append("You already have a reservation that coincides with this one. Please choose a different time.")
			else:
				policy_problems.append(f"{str(user)} already has a reservation that coincides with this one. Please choose a different time.")
		for area in new_reservation.area.get_ancestors(ascending=True, include_self=True):
			# Check reservations for all other children of the parent areas
			if not area.count_staff_in_occupancy:
				coincident_events = coincident_events.filter(user__is_staff=False)
			if not area.count_service_personnel_in_occupancy:
				coincident_events = coincident_events.filter(user__is_service_personnel=False)
			apply_to_user = (not user.is_staff and not user.is_service_personnel) or (user.is_staff and area.count_staff_in_occupancy) or (user.is_service_personnel and area.count_service_personnel_in_occupancy)
			if apply_to_user and area.maximum_capacity:
				children_events = coincident_events.filter(area_id__in=[area.id for area in area.get_descendants(include_self=True)])
				reservations = list(children_events)
				reservations.append(new_reservation)
				# Check only distinct users since the same user could make reservations in different rooms
				maximum_users, time = maximum_users_in_overlapping_reservations(reservations)
				if maximum_users > area.maximum_capacity:
					time_display = "at this time" if time is None else "at "+ format_datetime(time, "TIME_FORMAT")
					policy_problems.append(f"The {area} would be over its maximum capacity {time_display}. Please choose a different time.")

	# The user may not create, move, or resize a reservation to coincide with a scheduled outage.
	if new_reservation.reservation_item_type == ReservationItemType.TOOL:
		coincident_events = ScheduledOutage.objects.filter(
			Q(tool=new_reservation.tool) | Q(resource__fully_dependent_tools__in=[new_reservation.tool]))
	elif new_reservation.reservation_item_type == ReservationItemType.AREA:
		coincident_events = new_reservation.area.scheduled_outage_queryset()
	else:
		coincident_events = ScheduledOutage.objects.none()
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=new_reservation.start, end__lte=new_reservation.start)
	coincident_events = coincident_events.exclude(start__gte=new_reservation.end, end__gt=new_reservation.end)
	if coincident_events.count() > 0:
		policy_problems.append("Your reservation coincides with a scheduled outage. Please choose a different time.")


def should_enforce_policy(reservation: Reservation):
	""" Returns whether or not the policy rules should be enforced. """
	should_enforce = True

	item = reservation.reservation_item
	start_time = timezone.localtime(reservation.start)
	end_time = timezone.localtime(reservation.end)
	if item.policy_off_weekend and start_time.weekday() >= 5 and end_time.weekday() >= 5:
		should_enforce = False
	if item.policy_off_between_times and item.policy_off_start_time and item.policy_off_end_time:
		if item.policy_off_start_time <= item.policy_off_end_time:
			""" Range is something like 6am-6pm """
			if item.policy_off_start_time <= start_time.time() <= item.policy_off_end_time and item.policy_off_start_time <= end_time.time() <= item.policy_off_end_time:
				should_enforce = False
		else:
			""" Range is something like 6pm-6am """
			if (item.policy_off_start_time <= start_time.time() or start_time.time() <= item.policy_off_end_time) and (item.policy_off_start_time <= end_time.time() or end_time.time() <= item.policy_off_end_time):
				should_enforce = False
	return should_enforce


def check_policy_rules_for_item(user_creating_reservation: User, new_reservation: Reservation, cancelled_reservation: Optional[Reservation]):
	item_policy_problems = []
	# Calculate the duration of the reservation:
	duration = new_reservation.end - new_reservation.start

	# The reservation must be at least as long as the minimum block time for this item.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	item = new_reservation.reservation_item
	item_type = new_reservation.reservation_item_type
	if item.minimum_usage_block_time:
		minimum_block_time = timedelta(minutes=item.minimum_usage_block_time)
		if duration < minimum_block_time:
			item_policy_problems.append(f"Your reservation has a duration of {str(int(duration.total_seconds() / 60))} minutes. This {item_type.value} requires a minimum reservation duration of {str(int(minimum_block_time.total_seconds() / 60))} minutes.")

	# The reservation may not exceed the maximum block time for this tool.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item.maximum_usage_block_time:
		maximum_block_time = timedelta(minutes=item.maximum_usage_block_time)
		if duration > maximum_block_time:
			item_policy_problems.append(f"Your reservation has a duration of {str(int(duration.total_seconds() / 60))} minutes. Reservations for this {item_type.value} may not exceed {str(int(maximum_block_time.total_seconds() / 60))} minutes.")

	user = new_reservation.user

	# If there is a limit on number of reservations per user per day then verify that the user has not exceeded it.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item.maximum_reservations_per_day:
		start_of_day = new_reservation.start
		start_of_day = start_of_day.replace(hour=0, minute=0, second=0, microsecond=0)
		end_of_day = start_of_day + timedelta(days=1)
		reservations_for_that_day = Reservation.objects.filter(cancelled=False, shortened=False, start__gte=start_of_day, end__lte=end_of_day, user=user)
		reservations_for_that_day = reservations_for_that_day.filter(**new_reservation.reservation_item_filter)
		# Exclude any reservation that is being cancelled.
		if cancelled_reservation and cancelled_reservation.id:
			reservations_for_that_day = reservations_for_that_day.exclude(id=cancelled_reservation.id)
		if reservations_for_that_day.count() >= item.maximum_reservations_per_day:
			if user == user_creating_reservation:
				item_policy_problems.append(f"You may only have {str(item.maximum_reservations_per_day)} reservations for this {item_type.value} per day. Missed reservations are included when counting the number of reservations per day.")
			else:
				item_policy_problems.append(f"{str(user)} may only have {str(item.maximum_reservations_per_day)} reservations for this {item_type.value} per day. Missed reservations are included when counting the number of reservations per day.")

	# A minimum amount of time between reservations for the same user & same tool can be enforced.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item.minimum_time_between_reservations:
		buffer_time = timedelta(minutes=item.minimum_time_between_reservations)
		must_end_before = new_reservation.start - buffer_time
		too_close = Reservation.objects.filter(cancelled=False, shortened=False, user=user, end__gt=must_end_before, start__lt=new_reservation.start)
		too_close = too_close.filter(**new_reservation.reservation_item_filter)
		if cancelled_reservation and cancelled_reservation.id:
			too_close = too_close.exclude(id=cancelled_reservation.id)
		if too_close.exists():
			if user == user_creating_reservation:
				item_policy_problems.append(f"Separate reservations for this {item_type.value} that belong to you must be at least {str(item.minimum_time_between_reservations)} minutes apart from each other. The proposed reservation ends too close to another reservation.")
			else:
				item_policy_problems.append(f"Separate reservations for this {item_type.value} that belong to {str(user)} must be at least {str(item.minimum_time_between_reservations)} minutes apart from each other. The proposed reservation ends too close to another reservation.")
		must_start_after = new_reservation.end + buffer_time
		too_close = Reservation.objects.filter(cancelled=False, shortened=False, user=user, start__lt=must_start_after, end__gt=new_reservation.start)
		too_close = too_close.filter(**new_reservation.reservation_item_filter)
		if cancelled_reservation and cancelled_reservation.id:
			too_close = too_close.exclude(id=cancelled_reservation.id)
		if too_close.exists():
			if user == user_creating_reservation:
				item_policy_problems.append(f"Separate reservations for this {item_type.value} that belong to you must be at least {str(item.minimum_time_between_reservations)} minutes apart from each other. The proposed reservation begins too close to another reservation.")
			else:
				item_policy_problems.append(f"Separate reservations for this {item_type.value} that belong to {str(user)} must be at least {str(item.minimum_time_between_reservations)} minutes apart from each other. The proposed reservation begins too close to another reservation.")

	# Check that the user is not exceeding the maximum amount of time they may reserve in the future.
	# Staff may break this rule.
	# An explicit policy override allows this rule to be broken.
	if item.maximum_future_reservation_time:
		reservations_after_now = Reservation.objects.filter(cancelled=False, user=user, start__gte=timezone.now())
		reservations_after_now = reservations_after_now.filter(**new_reservation.reservation_item_filter)
		if cancelled_reservation and cancelled_reservation.id:
			reservations_after_now = reservations_after_now.exclude(id=cancelled_reservation.id)
		amount_reserved_in_the_future = new_reservation.duration()
		for r in reservations_after_now:
			amount_reserved_in_the_future += r.duration()
		if amount_reserved_in_the_future.total_seconds() / 60 > item.maximum_future_reservation_time:
			if user == user_creating_reservation:
				item_policy_problems.append(f"You may only reserve up to {str(item.maximum_future_reservation_time)} minutes of time on this {item_type.value}, starting from the current time onward.")
			else:
				item_policy_problems.append(f"{str(user)} may only reserve up to {str(item.maximum_future_reservation_time)} minutes of time on this {item_type.value}, starting from the current time onward.")

	return item_policy_problems


def check_policy_to_cancel_reservation(user_cancelling_reservation: User, reservation_to_cancel: Reservation, new_reservation: Optional[Reservation] = None):
	"""
	Checks the reservation deletion policy.
	If all checks pass the function returns an HTTP "OK" response.
	Otherwise, the function returns an HTTP "Bad Request" with an error message.
	"""

	move = new_reservation and new_reservation.start != reservation_to_cancel.start
	resize = new_reservation and new_reservation.start == reservation_to_cancel.start
	action = 'move' if move else 'resize' if resize else 'cancel'

	# Users may only cancel reservations that they own.
	# Staff may break this rule.
	if (reservation_to_cancel.user != user_cancelling_reservation) and not user_cancelling_reservation.is_staff:
		return HttpResponseBadRequest(f"You may not {action} reservations that you do not own.")

	# Users may not cancel reservations that have already ended.
	# Staff may break this rule.
	if reservation_to_cancel.end < timezone.now() and not user_cancelling_reservation.is_staff:
		return HttpResponseBadRequest(f"You may not {action} reservations that have already ended.")

	# Users may not cancel ongoing area reservations when they are currently logged in that area (unless they are extending it)
	# Staff may break this rule.
	if reservation_to_cancel.area and reservation_to_cancel.area.requires_reservation and not resize and reservation_to_cancel.start < timezone.now() < reservation_to_cancel.end and AreaAccessRecord.objects.filter(end=None, staff_charge=None, customer=reservation_to_cancel.user, area=reservation_to_cancel.area) and not user_cancelling_reservation.is_staff:
		if move:
			return HttpResponseBadRequest("You may only resize an area reservation while logged in that area.")
		else:
			return HttpResponseBadRequest("You may not cancel an area reservation while logged in that area.")

	if reservation_to_cancel.cancelled:
		return HttpResponseBadRequest("This reservation has already been cancelled by " + str(reservation_to_cancel.cancelled_by) + " on " + format_datetime(reservation_to_cancel.cancellation_time) + ".")

	if reservation_to_cancel.missed:
		return HttpResponseBadRequest("This reservation was missed and cannot be modified.")

	return HttpResponse()


def check_policy_to_create_outage(outage: ScheduledOutage):
	# Outages may not have a start time that is earlier than the end time.
	if outage.start >= outage.end:
		return "Outage start time (" + format_datetime(outage.start) + ") must be before the end time (" + format_datetime(outage.end) + ")."

	# The user may not create, move, or resize an outage to coincide with another user's reservation.
	coincident_events = Reservation.objects.filter(**outage.outage_item_filter).filter(cancelled=False, missed=False, shortened=False)
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	coincident_events = coincident_events.exclude(start__lt=outage.start, end__lte=outage.start)
	coincident_events = coincident_events.exclude(start__gte=outage.end, end__gt=outage.end)
	if coincident_events.count() > 0:
		return "Your scheduled outage coincides with a reservation that already exists. Please choose a different time."

	# No policy issues! The outage can be created...
	return None


def check_policy_to_enter_any_area(user: User):
	"""
	Checks the area access policy for a user.
	"""
	if not user.is_active:
		raise InactiveUserError(user=user)

	if user.active_project_count() < 1:
		raise NoActiveProjectsForUserError(user=user)

	if user.access_expiration is not None and user.access_expiration < date.today():
		raise PhysicalAccessExpiredUserError(user=user)

	user_has_access_to_at_least_one_area = user.accessible_access_levels().exists()
	if not user_has_access_to_at_least_one_area:
		raise NoPhysicalAccessUserError(user=user)


def check_policy_to_enter_this_area(area:Area, user:User):
	# If explicitly set on the Physical Access Level, staff may be exempt from being granted explicit access
	if user.is_staff and any([access_level.accessible() for access_level in PhysicalAccessLevel.objects.filter(allow_staff_access=True, area=area)]):
		pass
	else:
		# Check if the user normally has access to this area door at the current time (or access to any parent)
		if not any([access_level.accessible() for access_level in user.accessible_access_levels_for_area(area)]):
			first_closure_time = next(iter([access_level.ongoing_closure_time() for access_level in user.accessible_access_levels_for_area(area)]), None)
			raise NoAccessiblePhysicalAccessUserError(user=user, area=area, closure_time=first_closure_time)

	if not user.is_staff and not user.is_service_personnel:
		for a in area.get_ancestors(ascending=True, include_self=True):
			unavailable_resources = a.required_resources.filter(available=False)
			if unavailable_resources:
				raise UnavailableResourcesUserError(user=user, area=a, resources=unavailable_resources)

		# Non staff users may not enter an area during a scheduled outage.
		if area.scheduled_outage_in_progress():
			raise ScheduledOutageInProgressError(user=user, area=area)

		# If we reached maximum capacity, fail (only for non staff users)
		for a in area.get_ancestors(ascending=True, include_self=True):
			if a.maximum_capacity and 0 < a.maximum_capacity <= a.occupancy_count():
				raise MaximumCapacityReachedError(user=user, area=a)

		if area.requires_reservation and not area.get_current_reservation_for_user(user):
			raise ReservationRequiredUserError(user=user, area=area)


def check_billing_to_project(project: Project, user: User, item: Union[Tool, Area, Consumable, StaffCharge] = None):
	if project:
		if project not in user.active_projects():
			raise NotAllowedToChargeProjectException(project=project, user=user)

		if item:
			# Check if project only allows billing for certain tools
			allowed_tools = project.only_allow_tools.all()
			if allowed_tools.exists():
				if isinstance(item, Tool) and item not in allowed_tools:
					msg = f"{item.name} is not allowed for project {project.name}"
					raise ItemNotAllowedForProjectException(project, user, item.name, msg)
				elif isinstance(item, Area) and item.id not in distinct_qs_value_list(
						allowed_tools, '_requires_area_access_id'):
					msg = f"{item.name} is not allowed for project {project.name}"
					raise ItemNotAllowedForProjectException(project, user, item.name, msg)
			# Check if consumable withdrawals are allowed
			if isinstance(item, Consumable):
				if not project.allow_consumable_withdrawals:
					msg = f"Consumable withdrawals are not allowed for project {project.name}"
					raise ItemNotAllowedForProjectException(project, user, "Staff Charges", msg)


def maximum_users_in_overlapping_reservations(reservations: List[Reservation]) -> (int, datetime):
	"""
	Returns the maximum number of overlapping reservations and the earlier time the maximum is reached
	This will only count reservations made by different users. i.e. if a user has 3 reservations at the same
	time for different tools/areas, it will only count as one.
	"""
	# First we need to merge reservations by user, since one user could have more than one at the same time. (and we should only count it as one)
	intervals_by_user = defaultdict(list)
	for r in reservations:
		intervals_by_user[r.user.id].append((r.start, r.end))

	merged_intervals = []
	for user, intervals in intervals_by_user.items():
		merged_intervals.extend(recursive_merge(sorted(intervals).copy()))

	# Now let's count the maximum overlapping reservations
	times = []
	for interval in merged_intervals:
		start_time, end_time = interval[0], interval[1]
		times.append((start_time, 'start'))
		times.append((end_time, 'end'))
	times = sorted(times)

	count = 0
	max_count = 0
	max_time: Optional[datetime] = None
	for time in times:
		if time[1] == 'start':
			count += 1  # increment on arrival/start
		else:
			count -= 1  # decrement on departure/end
		# maintain maximum
		prev_count = max_count
		max_count = max(count, max_count)
		# maintain earlier time max is reached
		if max_count > prev_count:
			max_time = time[0]
	return max_count, max_time


def recursive_merge(intervals: List[tuple], start_index=0) -> List[tuple]:
	for i in range(start_index, len(intervals) - 1):
		if intervals[i][1] > intervals[i + 1][0]:
			new_start = intervals[i][0]
			new_end = intervals[i + 1][1]
			intervals[i] = (new_start, new_end)
			del intervals[i + 1]
			return recursive_merge(intervals.copy(), start_index=i)
	return intervals



import datetime
from datetime import timedelta

from django import template
from django.template import Context, Template
from django.template.defaultfilters import date, time
from django.urls import NoReverseMatch, reverse
from django.utils import timezone
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from pkg_resources import DistributionNotFound, get_distribution

from NEMO.views.customization import get_customization

register = template.Library()


@register.filter
def class_name(value):
	return value.__class__.__name__


@register.filter
def is_soon(time):
	""" 'Soon' is defined as within 10 minutes from now. """
	return time <= timezone.now() + timedelta(minutes=10)


@register.filter()
def to_int(value):
	return int(value)


@register.filter()
def to_date(value, arg=None):
	if value in (None, ''):
		return ''
	if isinstance(value, datetime.date):
		return date(value, arg)
	if isinstance(value, datetime.time):
		return time(value, arg)
	return value


@register.filter
def json_search_base(items_to_search, display = "__str__"):
	result = "["
	for item in items_to_search:
		attr = getattr(item, display, None)
		display_value = attr() if callable(attr) else attr
		result += '{{"name":"{0}", "id":{1}}},'.format(escape(display_value), item.id)
	result = result.rstrip(",") + "]"
	return mark_safe(result)


@register.simple_tag
def json_search_base_with_extra_fields(items_to_search, *extra_fields):
	"""
	This tag is similar to the json_search_base filter, but adds extra information upon request.
	The type of object is always provided in the JSON output. Thus, you have a heterogeneous collection
	of objects and differentiate them in your JavaScript. The extra fields are only added when an
	object actually has that attribute. Otherwise, the code skips over the request.
	"""
	result = "["
	for item in items_to_search:
		object_type = item.__class__.__name__.lower()
		result += '{{"name":"{0}", "id":"{1}", "type":"{2}"'.format(escape(str(item)), item.id, object_type)
		for x in extra_fields:
			if hasattr(item, x):
				result += ', "{0}":"{1}"'.format(x, getattr(item, x))
		result += "},"
	result = result.rstrip(",") + "]"
	return mark_safe(result)


@register.simple_tag
def navigation_url(url_name, description):
	try:
		return format_html('<li><a href="{}">{}</a></li>', reverse(url_name), description)
	except NoReverseMatch:
		return ""


@register.simple_tag
def res_question_tbody(dictionary):
	input_dict = dictionary[list(dictionary.keys())[0]]
	headers = list(input_dict.keys())
	header_cells = ''.join([format_html('<th>{}</th>', h) for h in headers])
	head_html = format_html('<thead><tr><th>#</th>{}</tr></thead>', mark_safe(header_cells))

	rows = []
	for i, (index, d) in enumerate(dictionary.items()):
		data_cells_html = ''.join([format_html("<td>{}</td>", d[h]) for h in headers])
		row_html = format_html('<tr><th>{}</th>{}</tr>', i + 1, mark_safe(data_cells_html))
		rows.append(row_html)
	body_html = format_html('<tbody>{}</tbody>', mark_safe(''.join(rows)))
	return head_html + body_html


@register.filter
def get_item(dictionary, key):
	return dictionary.get(key)


@register.simple_tag
def project_selection_display(project):
	project_selection_template = get_customization('project_selection_template')
	contents = "{{ project.name }}"
	try:
		contents = Template(project_selection_template).render(Context({'project': project}))
	except:
		pass
	return format_html(contents)


dist_version: str = "0"


@register.simple_tag()
def app_version() -> str:
	global dist_version
	if dist_version != "0":
		return dist_version
	else:
		try:
			dist_version = get_distribution("NEMO").version
		except DistributionNotFound:
			# package is not installed
			dist_version = None
			pass
	return dist_version



from logging import getLogger
from typing import List, Set

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.decorators.http import require_GET
from requests import get

from NEMO.decorators import staff_member_required
from NEMO.models import (
	Account,
	AreaAccessRecord,
	ConsumableWithdraw,
	Project,
	Reservation,
	StaffCharge,
	TrainingSession,
	UsageEvent,
	User,
)
from NEMO.utilities import (
	BasicDisplayTable,
	export_format_datetime,
	format_datetime,
	get_month_timeframe,
	month_list,
	parse_start_and_end_date,
)
from NEMO.views.api_billing import (
	BillableItem,
	billable_items_area_access_records,
	billable_items_consumable_withdrawals,
	billable_items_missed_reservations,
	billable_items_staff_charges,
	billable_items_training_sessions,
	billable_items_usage_events,
)

logger = getLogger(__name__)


# Class for Applications that can be used for autocomplete
class Application(object):
	def __init__(self, name):
		self.name = name
		self.id = name

	def __str__(self):
		return self.name


# We want to keep all the parameters of the request when switching tabs, so we are just replacing usage <-> billing urls
def get_url_for_other_tab(request):
	full_path_request = request.get_full_path()
	usage_url = reverse('usage')
	billing_url = reverse('billing')
	project_usage_url = reverse('project_usage')
	project_billing_url = reverse('project_billing')
	if project_usage_url in full_path_request:
		full_path_request = full_path_request.replace(project_usage_url, project_billing_url)
	elif project_billing_url in full_path_request:
		full_path_request = full_path_request.replace(project_billing_url, project_usage_url)
	elif usage_url in full_path_request:
		full_path_request = full_path_request.replace(usage_url, billing_url)
	elif billing_url in full_path_request:
		full_path_request = full_path_request.replace(billing_url, usage_url)
	return full_path_request


def get_project_applications():
	applications = []
	projects = Project.objects.filter(id__in=Project.objects.values('application_identifier').distinct().values_list('id', flat=True))
	for project in projects:
		if not any(list(filter(lambda app: app.name == project.application_identifier, applications))):
			applications.append(Application(project.application_identifier))
	return applications


def date_parameters_dictionary(request):
	if request.GET.get('start_date') and request.GET.get('end_date'):
		start_date, end_date = parse_start_and_end_date(request.GET.get('start_date'), request.GET.get('end_date'))
	else:
		start_date, end_date = get_month_timeframe()
	kind = request.GET.get("type")
	identifier = request.GET.get("id")
	dictionary = {
		'month_list': month_list(),
		'start_date': start_date,
		'end_date': end_date,
		'kind': kind,
		'identifier': identifier,
		'tab_url': get_url_for_other_tab(request),
		'billing_service': get_billing_service().get('available', False),
	}
	return dictionary, start_date, end_date, kind, identifier


@login_required
@require_GET
def usage(request):
	user: User = request.user
	user_managed_projects = get_managed_projects(user)
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	customer_filter = Q(customer=user) | Q(project__in=user_managed_projects)
	user_filter = Q(user=user) | Q(project__in=user_managed_projects)
	trainee_filter = Q(trainee=user) | Q(project__in=user_managed_projects)
	project_id = request.GET.get("pi_project")
	csv_export = bool(request.GET.get("csv", False))
	if user_managed_projects:
		base_dictionary['selected_project'] = "all"
	if project_id:
		project = get_object_or_404(Project, id=project_id)
		base_dictionary['selected_project'] = project
		customer_filter = customer_filter & Q(project=project)
		user_filter = user_filter & Q(project=project)
		trainee_filter = trainee_filter & Q(project=project)
	area_access = AreaAccessRecord.objects.filter(customer_filter).filter(end__gt=start_date, end__lte=end_date).order_by('-start')
	consumables = ConsumableWithdraw.objects.filter(customer_filter).filter(date__gt=start_date, date__lte=end_date)
	missed_reservations = Reservation.objects.filter(user_filter).filter(missed=True, end__gt=start_date, end__lte=end_date)
	staff_charges = StaffCharge.objects.filter(customer_filter).filter(end__gt=start_date, end__lte=end_date)
	training_sessions = TrainingSession.objects.filter(trainee_filter).filter(date__gt=start_date, date__lte=end_date)
	usage_events = UsageEvent.objects.filter(user_filter).filter(end__gt=start_date, end__lte=end_date)
	if csv_export:
		return csv_export_response(usage_events, area_access, training_sessions, staff_charges, consumables, missed_reservations)
	else:
		dictionary = {
			'area_access': area_access,
			'consumables': consumables,
			'missed_reservations': missed_reservations,
			'staff_charges': staff_charges,
			'training_sessions': training_sessions,
			'usage_events': usage_events,
			'can_export': True,
		}
		if user_managed_projects:
			dictionary['pi_projects'] = user_managed_projects
		dictionary['no_charges'] = not (dictionary['area_access'] or dictionary['consumables'] or dictionary['missed_reservations'] or dictionary['staff_charges'] or dictionary['training_sessions'] or dictionary['usage_events'])
		return render(request, 'usage/usage.html', {**base_dictionary, **dictionary})


@login_required
@require_GET
def billing(request):
	user: User = request.user
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	if not base_dictionary['billing_service']:
		return redirect('usage')
	user_project_applications = list(user.active_projects().values_list('application_identifier', flat=True)) + list(user.managed_projects.values_list('application_identifier', flat=True))
	formatted_applications = ','.join(map(str, set(user_project_applications)))
	try:
		billing_dictionary = billing_dict(start_date, end_date, user, formatted_applications)
		return render(request, 'usage/billing.html', {**base_dictionary, **billing_dictionary})
	except Exception as e:
		logger.warning(str(e))
		return render(request, 'usage/billing.html', base_dictionary)


@staff_member_required
@require_GET
def project_usage(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)

	area_access, consumables, missed_reservations, staff_charges, training_sessions, usage_events = None, None, None, None, None, None

	projects = []
	user = None
	selection = ''
	try:
		if kind == 'application':
			projects = Project.objects.filter(application_identifier=identifier)
			selection = identifier
		elif kind == 'project':
			projects = [Project.objects.get(id=identifier)]
			selection = projects[0].name
		elif kind == 'account':
			account = Account.objects.get(id=identifier)
			projects = Project.objects.filter(account=account)
			selection = account.name
		elif kind == 'user':
			user = User.objects.get(id=identifier)
			projects = user.active_projects()
			selection = str(user)

		if projects:
			area_access = AreaAccessRecord.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date).order_by('-start')
			consumables = ConsumableWithdraw.objects.filter(project__in=projects, date__gt=start_date, date__lte=end_date)
			missed_reservations = Reservation.objects.filter(project__in=projects, missed=True, end__gt=start_date, end__lte=end_date)
			staff_charges = StaffCharge.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date)
			training_sessions = TrainingSession.objects.filter(project__in=projects, date__gt=start_date, date__lte=end_date)
			usage_events = UsageEvent.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date)
			if user:
				area_access = area_access.filter(customer=user)
				consumables = consumables.filter(customer=user)
				missed_reservations = missed_reservations.filter(user=user)
				staff_charges = staff_charges.filter(customer=user)
				training_sessions = training_sessions.filter(trainee=user)
				usage_events = usage_events.filter(user=user)
			if bool(request.GET.get("csv", False)):
				return csv_export_response(usage_events, area_access, training_sessions, staff_charges, consumables, missed_reservations)
	except:
		pass
	dictionary = {
		'search_items': set(Account.objects.all()) | set(Project.objects.all()) | set(get_project_applications()) | set(User.objects.filter(is_active=True)),
		'area_access': area_access,
		'consumables': consumables,
		'missed_reservations': missed_reservations,
		'staff_charges': staff_charges,
		'training_sessions': training_sessions,
		'usage_events': usage_events,
		'project_autocomplete': True,
		'selection': selection,
		'can_export': True,
	}
	dictionary['no_charges'] = not (dictionary['area_access'] or dictionary['consumables'] or dictionary['missed_reservations'] or dictionary['staff_charges'] or dictionary['training_sessions'] or dictionary['usage_events'])
	return render(request, 'usage/usage.html', {**base_dictionary, **dictionary})


@staff_member_required
@require_GET
def project_billing(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	if not base_dictionary['billing_service']:
		return redirect('project_usage')
	base_dictionary['project_autocomplete'] = True
	base_dictionary['search_items'] = set(Account.objects.all()) | set(Project.objects.all()) | set(get_project_applications()) | set(User.objects.filter(is_active=True))

	project_id = None
	account_id = None
	user = None
	formatted_applications = None
	selection = ''
	try:
		if kind == 'application':
			formatted_applications = identifier
			selection = identifier
		elif kind == 'project':
			projects = [Project.objects.get(id=identifier)]
			formatted_applications = projects[0].application_identifier
			project_id = identifier
			selection = projects[0].name
		elif kind == 'account':
			account = Account.objects.get(id=identifier)
			projects = Project.objects.filter(account=account, active=True, account__active=True)
			formatted_applications = ','.join(map(str, set(projects.values_list('application_identifier', flat=True)))) if projects else None
			account_id = account.id
			selection = account.name
		elif kind == 'user':
			user = User.objects.get(id=identifier)
			projects = user.active_projects()
			formatted_applications = ','.join(map(str, set(projects.values_list('application_identifier', flat=True)))) if projects else None
			selection = str(user)

		base_dictionary['selection'] = selection
		billing_dictionary = billing_dict(start_date, end_date, user, formatted_applications, project_id, account_id=account_id, force_pi=True if not user else False)
		return render(request, 'usage/billing.html', {**base_dictionary, **billing_dictionary})
	except Exception as e:
		logger.warning(str(e))
		return render(request, 'usage/billing.html', base_dictionary)


def is_user_pi(user: User, latest_pis_data, activity, user_managed_applications: List[str]):
	# Check if the user is set as a PI in NEMO, otherwise check from latest_pis_data
	application = activity['application_name']
	if application in user_managed_applications:
		return True
	else:
		application_pi_row = next((x for x in latest_pis_data if x['application_name'] == application), None)
		return application_pi_row is not None and (user.username == application_pi_row['username'] or (user.first_name == application_pi_row['first_name'] and user.last_name == application_pi_row['last_name']))


def billing_dict(start_date, end_date, user, formatted_applications, project_id=None, account_id=None, force_pi=False):
	# The parameter force_pi allows us to display information as if the user was the project pi
	# This is useful on the admin project billing page tp display other project users for example
	dictionary = {}

	billing_service = get_billing_service()
	if not billing_service.get('available', False):
		return dictionary

	cost_activity_url = billing_service['cost_activity_url']
	project_lead_url = billing_service['project_lead_url']
	keyword_arguments = billing_service['keyword_arguments']

	cost_activity_params = {
		'created_date_gte': f"'{start_date.strftime('%m/%d/%Y')}'",
		'created_date_lt': f"'{end_date.strftime('%m/%d/%Y')}'",
		'application_names': f"'{formatted_applications}'",
		'$format': 'json'
	}
	cost_activity_response = get(cost_activity_url, params=cost_activity_params, **keyword_arguments)
	cost_activity_data = cost_activity_response.json()['d']

	if not force_pi:
		latest_pis_params = {'$format': 'json'}
		latest_pis_response = get(project_lead_url, params=latest_pis_params, **keyword_arguments)
		latest_pis_data = latest_pis_response.json()['d']

	project_totals = {}
	application_totals = {}
	account_totals = {}
	user_pi_applications = list()
	# Construct a tree of account, application, project, and member total spending
	cost_activities_tree = {}
	user_managed_applications = [project.application_identifier for project in user.managed_projects.all()]
	for activity in cost_activity_data:
		if (project_id and activity['project_id'] != str(project_id)) or (account_id and activity['account_id'] != str(account_id)):
			continue
		project_totals.setdefault(activity['project_id'], 0)
		application_totals.setdefault(activity['application_id'], 0)
		account_totals.setdefault(activity['account_id'], 0)
		account_key = (activity['account_id'], activity['account_name'])
		application_key = (activity['application_id'], activity['application_name'])
		project_key = (activity['project_id'], activity['project_name'])
		user_key = (activity['member_id'], User.objects.filter(id__in=[activity['member_id']]).first())
		user_is_pi = is_user_pi(user, latest_pis_data, activity, user_managed_applications) if not force_pi else True
		if user_is_pi:
			user_pi_applications.append(activity['application_id'])
		if user_is_pi or str(user.id) == activity['member_id']:
			cost_activities_tree.setdefault((activity['account_id'], activity['account_name']), {})
			cost_activities_tree[account_key].setdefault(application_key, {})
			cost_activities_tree[account_key][application_key].setdefault(project_key, {})
			cost_activities_tree[account_key][application_key][project_key].setdefault(user_key, 0)
			cost = 0
			if activity['cost'] is not None:
				cost = -activity['cost'] if activity['activity_type'] == 'refund_activity' else activity['cost']
			cost_activities_tree[account_key][application_key][project_key][user_key] = cost_activities_tree[account_key][application_key][project_key][user_key] + cost
			project_totals[activity['project_id']] = project_totals[activity['project_id']] + cost
			application_totals[activity['application_id']] = application_totals[activity['application_id']] + cost
			account_totals[activity['account_id']] = account_totals[activity['account_id']] + cost
	dictionary['spending'] = {
		'activities': cost_activities_tree,
		'project_totals': project_totals,
		'application_totals': application_totals,
		'account_totals': account_totals,
		'user_pi_applications': user_pi_applications
	} if cost_activities_tree else {'activities': {}}
	return dictionary


def csv_export_response(usage_events, area_access, training_sessions, staff_charges, consumables, missed_reservations):
	table_result = BasicDisplayTable()
	table_result.add_header(("type", "Type"))
	table_result.add_header(("user", "User"))
	table_result.add_header(("name", "Item"))
	table_result.add_header(("details", "Details"))
	table_result.add_header(("project", "Project"))
	table_result.add_header(("start", "Start time"))
	table_result.add_header(("end", "End time"))
	table_result.add_header(("quantity", "Quantity"))
	data: List[BillableItem] = []
	data.extend(billable_items_missed_reservations(missed_reservations))
	data.extend(billable_items_consumable_withdrawals(consumables))
	data.extend(billable_items_staff_charges(staff_charges))
	data.extend(billable_items_training_sessions(training_sessions))
	data.extend(billable_items_area_access_records(area_access))
	data.extend(billable_items_usage_events(usage_events))
	for billable_item in data:
		# Format start & end times
		billable_item.start = format_datetime(billable_item.start, "SHORT_DATETIME_FORMAT")
		billable_item.end = format_datetime(billable_item.end, "SHORT_DATETIME_FORMAT")
		table_result.add_row(vars(billable_item))
	response = table_result.to_csv()
	filename = f"usage_export_{export_format_datetime()}.csv"
	response["Content-Disposition"] = f'attachment; filename="{filename}"'
	return response


def get_managed_projects(user: User) -> Set[Project]:
	# This function will get managed projects from NEMO and also attempt to get them from billing service
	managed_projects = set(list(user.managed_projects.all()))
	billing_service = get_billing_service()
	if billing_service.get('available', False):
		# if we have a billing service, use it to determine project lead
		project_lead_url = billing_service['project_lead_url']
		keyword_arguments = billing_service['keyword_arguments']
		latest_pis_params = {'$format': 'json'}
		latest_pis_response = get(project_lead_url, params=latest_pis_params, **keyword_arguments)
		latest_pis_data = latest_pis_response.json()['d']
		for project_lead in latest_pis_data:
			if project_lead['username'] == user.username or (
					project_lead['first_name'] == user.first_name and project_lead['last_name'] == user.last_name):
				try:
					for managed_project in Project.objects.filter(application_identifier=project_lead['application_name']):
						managed_projects.add(managed_project)
				except Project.DoesNotExist:
					pass
	return managed_projects


def get_billing_service():
	return getattr(settings, 'BILLING_SERVICE', {})



from django.core.management import BaseCommand

from NEMO.views.calendar import do_cancel_unused_reservations


class Command(BaseCommand):
	help = (
		"Run every minute to cancel unused reservations and mark them as missed. "
		"Only applicable to areas or tools having a missed reservation threshold value."
	)

	def handle(self, *args, **options):
		do_cancel_unused_reservations()



from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.template import Template, RequestContext, Context
from django.views.decorators.http import require_http_methods

from NEMO.models import User, Project
from NEMO.utilities import send_mail, EmailCategory
from NEMO.views.customization import get_customization, get_media_file_contents


@login_required
@require_http_methods(["GET", "POST"])
def facility_rules(request):
	if request.method == "GET":
		tutorial = get_media_file_contents("facility_rules_tutorial.html")
		if tutorial:
			dictionary = {
				"active_user_count": User.objects.filter(is_active=True).count(),
				"active_project_count": Project.objects.filter(active=True).count(),
			}
			tutorial = Template(tutorial).render(RequestContext(request, dictionary))
		return render(request, "facility_rules.html", {"facility_rules_tutorial": tutorial})
	elif request.method == "POST":
		facility_name = get_customization("facility_name")
		summary = request.POST.get("making_reservations_summary", "").strip()[:3000]
		dictionary = {"user": request.user, "making_reservations_rule_summary": summary}
		abuse_email = get_customization("abuse_email_address")
		email_contents = get_media_file_contents("facility_rules_tutorial_email.html")
		if abuse_email and email_contents:
			message = Template(email_contents, dictionary).render(Context(dictionary))
			send_mail(
				subject=f"{facility_name} rules tutorial",
				content=message,
				from_email=abuse_email,
				to=[abuse_email],
				email_category=EmailCategory.SYSTEM,
			)
		dictionary = {
			"title": f"{facility_name} rules tutorial",
			"heading": "Tutorial complete!",
			"content": "Tool usage and reservation privileges have been enabled on your user account.",
		}
		request.user.training_required = False
		request.user.save()
		return render(request, "acknowledgement.html", dictionary)



from datetime import datetime, timedelta
from http import HTTPStatus
from itertools import chain
from json import JSONDecodeError, loads
from logging import getLogger
from typing import Dict, List

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseNotFound, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.template import Context, Template
from django.template.defaultfilters import linebreaksbr
from django.utils import formats, timezone
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_POST

from NEMO import rates
from NEMO.decorators import staff_member_required, synchronized
from NEMO.exceptions import RequiredUnansweredQuestionsException
from NEMO.forms import CommentForm, nice_errors
from NEMO.models import (
	AreaAccessRecord,
	Comment,
	Configuration,
	ConfigurationHistory,
	Project,
	Reservation,
	StaffCharge,
	Task,
	TaskCategory,
	TaskStatus,
	Tool,
	ToolUsageCounter,
	UsageEvent,
	User,
)
from NEMO.utilities import (
	BasicDisplayTable,
	EmailCategory,
	beginning_of_the_day,
	end_of_the_day,
	export_format_datetime,
	extract_times,
	format_datetime,
	quiet_int,
	send_mail,
)
from NEMO.views.calendar import shorten_reservation
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.policy import check_policy_to_disable_tool, check_policy_to_enable_tool
from NEMO.widgets.configuration_editor import ConfigurationEditor
from NEMO.widgets.dynamic_form import DynamicForm, PostUsageQuestion, render_group_questions
from NEMO.widgets.item_tree import ItemTree

tool_control_logger = getLogger(__name__)


@login_required
@require_GET
def tool_control(request, item_type="tool", tool_id=None):
	# item_type is needed for compatibility with 'view_calendar' view on mobile
	""" Presents the tool control view to the user, allowing them to begin/end using a tool or see who else is using it. """
	user: User = request.user
	if user.active_project_count() == 0:
		return render(request, "no_project.html")
	# The tool-choice sidebar is not available for mobile devices, so redirect the user to choose a tool to view.
	if request.device == "mobile" and tool_id is None:
		return redirect("choose_item", next_page="tool_control")
	tools = Tool.objects.filter(visible=True).order_by("_category", "name")
	dictionary = {"tools": tools, "selected_tool": tool_id}
	# The tool-choice sidebar only needs to be rendered for desktop devices, not mobile devices.
	if request.device == "desktop":
		dictionary["rendered_item_tree_html"] = ItemTree().render(None, {"tools": tools, "user": user})
	return render(request, "tool_control/tool_control.html", dictionary)


@login_required
@require_GET
def tool_status(request, tool_id):
	""" Gets the current status of the tool (that is, whether it is currently in use or not). """
	tool = get_object_or_404(Tool, id=tool_id, visible=True)

	dictionary = {
		"tool": tool,
		"tool_rate": rates.rate_class.get_tool_rate(tool),
		"task_categories": TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT),
		"rendered_configuration_html": tool.configuration_widget(request.user),
		"mobile": request.device == "mobile",
		"task_statuses": TaskStatus.objects.all(),
		"post_usage_questions": DynamicForm(tool.post_usage_questions).render("tool_usage_group_question", tool_id),
		"configs": get_tool_full_config_history(tool),
	}

	try:
		current_reservation = Reservation.objects.get(
			start__lt=timezone.now(),
			end__gt=timezone.now(),
			cancelled=False,
			missed=False,
			shortened=False,
			user=request.user,
			tool=tool,
		)
		if request.user == current_reservation.user:
			dictionary["time_left"] = current_reservation.end
	except Reservation.DoesNotExist:
		pass

	# Staff need the user list to be able to qualify users for the tool.
	if request.user.is_staff:
		dictionary["users"] = User.objects.filter(is_active=True)

	return render(request, "tool_control/tool_status.html", dictionary)


@staff_member_required
@require_GET
def use_tool_for_other(request):
	dictionary = {"users": User.objects.filter(is_active=True).exclude(id=request.user.id)}
	return render(request, "tool_control/use_tool_for_other.html", dictionary)


def get_tool_full_config_history(tool: Tool):
	# tool config by user and tool and time
	configs = []
	config_history = ConfigurationHistory.objects.filter(configuration__tool_id=tool.id).order_by("-modification_time")[
					 :20
					 ]
	configurations = tool.current_ordered_configurations()
	for c in config_history:
		for co in configurations:
			if co == c.configuration:
				current_settings = co.current_settings_as_list()
				current_settings[c.slot] = c.setting
				co.current_settings = ", ".join(current_settings)
		config_input = {"configurations": configurations, "render_as_form": False}
		configuration = ConfigurationEditor()
		configs.append(
			{"modification_time": c.modification_time, "user": c.user, "html": configuration.render(None, config_input)}
		)
	return configs


@login_required
@require_POST
def usage_data_history(request, tool_id):
	""" This method return a dictionary of headers and rows containing run_data information for Usage Events """
	csv_export = bool(request.POST.get("csv", False))
	start, end = extract_times(request.POST, start_required=False, end_required=False)
	last = request.POST.get("data_history_last")
	user_id = request.POST.get("data_history_user_id")
	show_project_info = request.POST.get("show_project_info")
	if not last and not start and not end:
		# Default to last 25 records
		last = 25
	usage_events = UsageEvent.objects.filter(tool_id=tool_id, end__isnull=False).order_by("-end")
	if start:
		usage_events = usage_events.filter(end__gte=beginning_of_the_day(start))
	if end:
		usage_events = usage_events.filter(end__lte=end_of_the_day(end))
	if user_id:
		try:
			usage_events = usage_events.filter(user_id=int(user_id))
		except ValueError:
			pass
	if last:
		try:
			last = int(last)
		except ValueError:
			last = 25
		usage_events = usage_events[:last]
	table_result = BasicDisplayTable()
	table_result.add_header(("user", "User"))
	if show_project_info:
		table_result.add_header(("project", "Project"))
	table_result.add_header(("date", "Date"))
	for usage_event in usage_events:
		if usage_event.run_data:
			usage_data = {}
			try:
				user_data = f"{usage_event.user.first_name} {usage_event.user.last_name}"
				date_data = format_datetime(usage_event.end, "SHORT_DATETIME_FORMAT")
				run_data: Dict = loads(usage_event.run_data)
				for question_key, question in run_data.items():
					if "user_input" in question:
						if question["type"] == "group":
							for sub_question in question["questions"]:
								table_result.add_header((sub_question["name"], sub_question["title"]))
							for index, user_inputs in question["user_input"].items():
								if index == "0":
									# Special case here the "initial" group of user inputs will go along with the rest of the non-group user inputs
									for name, user_input in user_inputs.items():
										usage_data[name] = user_input
								else:
									# For the other groups of user inputs, we have to add a whole new row
									group_usage_data = {}
									for name, user_input in user_inputs.items():
										group_usage_data[name] = user_input
									if group_usage_data:
										group_usage_data["user"] = user_data
										group_usage_data["date"] = date_data
										if show_project_info:
											group_usage_data["project"] = usage_event.project.name
										table_result.add_row(group_usage_data)
						else:
							table_result.add_header((question_key, question["title"]))
							usage_data[question_key] = question["user_input"]
				if usage_data:
					usage_data["user"] = user_data
					usage_data["date"] = date_data
					if show_project_info:
						usage_data["project"] = usage_event.project.name
					table_result.add_row(usage_data)
			except JSONDecodeError:
				tool_control_logger.debug("error decoding run_data: " + usage_event.run_data)
	if csv_export:
		response = table_result.to_csv()
		filename = f"tool_usage_data_export_{export_format_datetime()}.csv"
		response["Content-Disposition"] = f'attachment; filename="{filename}"'
		return response
	else:
		dictionary = {
			"tool_id": tool_id,
			"data_history_start": start,
			"data_history_end": end,
			"data_history_last": str(last),
			"usage_data_table": table_result,
			"data_history_user": User.objects.get(id=user_id) if user_id else None,
			"show_project_info": show_project_info or False,
			"users": User.objects.filter(is_active=True)
		}
		return render(request, "tool_control/usage_data.html", dictionary)


@login_required
@require_POST
def tool_configuration(request):
	""" Sets the current configuration of a tool. """
	try:
		configuration = Configuration.objects.get(id=request.POST["configuration_id"])
	except:
		return HttpResponseNotFound("Configuration not found.")
	if configuration.tool.in_use():
		return HttpResponseBadRequest("Cannot change a configuration while a tool is in use.")
	if not configuration.user_is_maintainer(request.user):
		return HttpResponseBadRequest("You are not authorized to change this configuration.")
	try:
		slot = int(request.POST["slot"])
		choice = int(request.POST["choice"])
	except:
		return HttpResponseBadRequest("Invalid configuration parameters.")
	try:
		configuration.replace_current_setting(slot, choice)
	except IndexError:
		return HttpResponseBadRequest("Invalid configuration choice.")
	configuration.save()
	history = ConfigurationHistory()
	history.configuration = configuration
	history.slot = slot
	history.user = request.user
	history.setting = configuration.get_current_setting(slot)
	history.save()
	return HttpResponse()


@login_required
@require_POST
def create_comment(request):
	form = CommentForm(request.POST)
	if not form.is_valid():
		return HttpResponseBadRequest(nice_errors(form).as_ul())
	comment = form.save(commit=False)
	comment.content = comment.content.strip()
	comment.author = request.user
	comment.expiration_date = (
		None
		if form.cleaned_data["expiration"] == 0
		else timezone.now() + timedelta(days=form.cleaned_data["expiration"])
	)
	comment.save()
	return redirect("tool_control")


@login_required
@require_POST
def hide_comment(request, comment_id):
	comment = get_object_or_404(Comment, id=comment_id)
	if comment.author_id != request.user.id and not request.user.is_staff:
		return HttpResponseBadRequest("You may only hide a comment if you are its author or a staff member.")
	comment.visible = False
	comment.hidden_by = request.user
	comment.hide_date = timezone.now()
	comment.save()
	return redirect("tool_control")


def determine_tool_status(tool):
	# Make the tool operational when all problems are resolved that require a shutdown.
	if tool.task_set.filter(force_shutdown=True, cancelled=False, resolved=False).count() == 0:
		tool.operational = True
	else:
		tool.operational = False
	tool.save()


@login_required
@require_POST
@synchronized('tool_id')
def enable_tool(request, tool_id, user_id, project_id, staff_charge):
	""" Enable a tool for a user. The user must be qualified to do so based on the lab usage policy. """

	if not settings.ALLOW_CONDITIONAL_URLS:
		return HttpResponseBadRequest(
			"Tool control is only available on campus. We're working to change that! Thanks for your patience."
		)

	tool = get_object_or_404(Tool, id=tool_id)
	operator = request.user
	user = get_object_or_404(User, id=user_id)
	project = get_object_or_404(Project, id=project_id)
	staff_charge = staff_charge == "true"
	bypass_interlock = request.POST.get("bypass", 'False') == 'True'
	response = check_policy_to_enable_tool(tool, operator, user, project, staff_charge)
	if response.status_code != HTTPStatus.OK:
		return response

	# All policy checks passed so enable the tool for the user.
	if tool.interlock and not tool.interlock.unlock():
		if bypass_interlock and interlock_bypass_allowed(user):
			pass
		else:
			return interlock_error("Enable", user)

	# Start staff charge before tool usage
	if staff_charge:
		new_staff_charge = StaffCharge()
		new_staff_charge.staff_member = request.user
		new_staff_charge.customer = user
		new_staff_charge.project = project
		new_staff_charge.save()
		# If the tool requires area access, start charging area access time
		if tool.requires_area_access:
			area_access = AreaAccessRecord()
			area_access.area = tool.requires_area_access
			area_access.staff_charge = new_staff_charge
			area_access.customer = new_staff_charge.customer
			area_access.project = new_staff_charge.project
			area_access.save()

	# Create a new usage event to track how long the user uses the tool.
	new_usage_event = UsageEvent()
	new_usage_event.operator = operator
	new_usage_event.user = user
	new_usage_event.project = project
	new_usage_event.tool = tool
	new_usage_event.save()

	return response


@login_required
@require_POST
@synchronized('tool_id')
def disable_tool(request, tool_id):
	if not settings.ALLOW_CONDITIONAL_URLS:
		return HttpResponseBadRequest("Tool control is only available on campus.")

	tool = get_object_or_404(Tool, id=tool_id)
	if tool.get_current_usage_event() is None:
		return HttpResponse()
	downtime = timedelta(minutes=quiet_int(request.POST.get("downtime")))
	bypass_interlock = request.POST.get("bypass", 'False') == 'True'
	response = check_policy_to_disable_tool(tool, request.user, downtime)
	if response.status_code != HTTPStatus.OK:
		return response

	# All policy checks passed so disable the tool for the user.
	if tool.interlock and not tool.interlock.lock():
		if bypass_interlock and interlock_bypass_allowed(request.user):
			pass
		else:
			return interlock_error("Disable", request.user)

	# Shorten the user's tool reservation since we are now done using the tool
	shorten_reservation(user=request.user, item=tool, new_end=timezone.now() + downtime)

	# End the current usage event for the tool
	current_usage_event = tool.get_current_usage_event()
	current_usage_event.end = timezone.now() + downtime

	# Collect post-usage questions
	dynamic_form = DynamicForm(tool.post_usage_questions)

	try:
		current_usage_event.run_data = dynamic_form.extract(request)
	except RequiredUnansweredQuestionsException as e:
		if request.user.is_staff and request.user != current_usage_event.operator and current_usage_event.user != request.user:
			# if a staff is forcing somebody off the tool and there are required questions, send an email and proceed
			current_usage_event.run_data = e.run_data
			email_managers_required_questions_disable_tool(current_usage_event.operator, request.user, tool, e.questions)
		else:
			return HttpResponseBadRequest(str(e))

	dynamic_form.charge_for_consumables(
		current_usage_event.user,
		current_usage_event.operator,
		current_usage_event.project,
		current_usage_event.run_data,
		request
	)
	dynamic_form.update_tool_counters(current_usage_event.run_data, tool.id)

	current_usage_event.save()
	user: User = request.user
	if user.charging_staff_time():
		existing_staff_charge = user.get_staff_charge()
		if (
				existing_staff_charge.customer == current_usage_event.user
				and existing_staff_charge.project == current_usage_event.project
		):
			response = render(request, "staff_charges/reminder.html", {"tool": tool})

	return response


@login_required
@require_GET
def past_comments_and_tasks(request):
	start, end = extract_times(request.GET, start_required=False, end_required=False)
	search = request.GET.get("search")
	if not start and not end and not search:
		return HttpResponseBadRequest("Please enter a search keyword, start date or end date.")
	tool_id = request.GET.get("tool_id")
	try:
		tasks = Task.objects.filter(tool_id=tool_id)
		comments = Comment.objects.filter(tool_id=tool_id, staff_only=False)
		if start:
			tasks = tasks.filter(creation_time__gt=start)
			comments = comments.filter(creation_date__gt=start)
		if end:
			tasks = tasks.filter(creation_time__lt=end)
			comments = comments.filter(creation_date__lt=end)
		if search:
			tasks = tasks.filter(problem_description__icontains=search)
			comments = comments.filter(content__icontains=search)
	except:
		return HttpResponseBadRequest("Task and comment lookup failed.")
	past = list(chain(tasks, comments))
	past.sort(key=lambda x: getattr(x, "creation_time", None) or getattr(x, "creation_date", None))
	past.reverse()
	if request.GET.get("export"):
		return export_comments_and_tasks_to_text(past)
	return render(request, "tool_control/past_tasks_and_comments.html", {"past": past})


@login_required
@require_GET
def ten_most_recent_past_comments_and_tasks(request, tool_id):
	tasks = Task.objects.filter(tool_id=tool_id).order_by("-creation_time")[:10]
	comments = Comment.objects.filter(tool_id=tool_id, staff_only=False).order_by("-creation_date")[:10]
	past = list(chain(tasks, comments))
	past.sort(key=lambda x: getattr(x, "creation_time", None) or getattr(x, "creation_date", None))
	past.reverse()
	past = past[0:10]
	if request.GET.get("export"):
		return export_comments_and_tasks_to_text(past)
	return render(request, "tool_control/past_tasks_and_comments.html", {"past": past})


def export_comments_and_tasks_to_text(comments_and_tasks: List):
	content = "No tasks or comments were created between these dates." if not comments_and_tasks else ""
	for item in comments_and_tasks:
		if isinstance(item, Comment):
			comment: Comment = item
			staff_only = "staff only " if comment.staff_only else ""
			content += f"On {format_datetime(comment.creation_date)} {comment.author} wrote this {staff_only}comment:\n"
			content += f"{comment.content}\n"
			if comment.hide_date:
				content += f"{comment.hidden_by} hid the comment on {format_datetime(comment.hide_date)}.\n"
		elif isinstance(item, Task):
			task: Task = item
			content += f"On {format_datetime(task.creation_time)} {task.creator} created this task:\n"
			if task.problem_category:
				content += f"{task.problem_category.name}\n"
			if task.force_shutdown:
				content += "\nThe tool was shut down because of this task.\n"
			if task.progress_description:
				content += f"\n{task.progress_description}\n"
			if task.resolved:
				resolution_category = f"({task.resolution_category}) " if task.resolution_category else ""
				content += f"\nResolved {resolution_category}On {format_datetime(task.resolution_time)} by {task.resolver }.\n"
				if task.resolution_description:
					content += f"{task.resolution_description}\n"
			elif task.cancelled:
				content += f"\nCancelled On {format_datetime(task.resolution_time)} by {task.resolver}.\n"
		content += "\n---------------------------------------------------\n\n"
	response = HttpResponse(content, content_type='text/plain')
	response['Content-Disposition'] = 'attachment; filename={0}'.format(f"comments_and_tasks_export_{export_format_datetime()}.txt")
	return response


@login_required
@require_GET
def tool_usage_group_question(request, tool_id, group_name):
	tool = get_object_or_404(Tool, id=tool_id)
	return HttpResponse(render_group_questions(request, tool.post_usage_questions, "tool_usage_group_question", tool_id, group_name))


@staff_member_required
@require_GET
def reset_tool_counter(request, counter_id):
	counter = get_object_or_404(ToolUsageCounter, id=counter_id)
	counter.last_reset_value = counter.value
	counter.value = 0
	counter.last_reset = datetime.now()
	counter.last_reset_by = request.user
	counter.save()

	# Save a comment about the counter being reset.
	comment = Comment()
	comment.tool = counter.tool
	comment.content = f"The {counter.name} counter was reset to 0. Its last value was {counter.last_reset_value}."
	comment.author = request.user
	comment.expiration_date = datetime.now() + timedelta(weeks=1)
	comment.save()

	# Send an email to Lab Managers about the counter being reset.
	facility_managers = User.objects.filter(is_active=True, is_facility_manager=True).values_list('email', flat=True)
	if facility_managers:
		message = f"""The {counter.name} counter for the {counter.tool.name} was reset to 0 on {formats.localize(counter.last_reset)} by {counter.last_reset_by}.
	
Its last value was {counter.last_reset_value}."""
		send_mail(
			subject=f"{counter.tool.name} counter reset",
			content=message,
			from_email=settings.SERVER_EMAIL,
			to=facility_managers,
			email_category=EmailCategory.SYSTEM,
		)
	return redirect("tool_control")


def interlock_bypass_allowed(user: User):
	return user.is_staff or get_customization('allow_bypass_interlock_on_failure') == 'enabled'


def interlock_error(action:str, user:User):
	error_message = get_customization('tool_interlock_failure_message')
	dictionary = {
		"message": linebreaksbr(error_message),
		"bypass_allowed": interlock_bypass_allowed(user),
		"action": action
	}
	return JsonResponse(dictionary, status=501)


def email_managers_required_questions_disable_tool(tool_user:User, staff_member:User, tool:Tool, questions:List[PostUsageQuestion]):
	abuse_email_address = get_customization('abuse_email_address')
	facility_managers = User.objects.filter(is_active=True, is_facility_manager=True).values_list('email', flat=True)
	ccs = set(tuple([r for r in [staff_member.email, tool.primary_owner.email, *tool.backup_owners.all().values_list('email', flat=True), *facility_managers] if r]))
	display_questions = "".join([linebreaksbr(mark_safe(question.render_as_text())) + "<br/><br/>" for question in questions])
	message = f"""
Dear {tool_user.get_name()},<br/>
You have been logged off by staff from the {tool} that requires answers to the following post-usage questions:<br/>
<br/>
{display_questions}
<br/>
Regards,<br/>
<br/>
NanoFab Management<br/>
"""
	send_mail(subject=f"Unanswered post‑usage questions after logoff from the {tool.name}", content=message, from_email=abuse_email_address, to=[tool_user.email], cc=ccs, email_category=EmailCategory.ABUSE)


def send_tool_usage_counter_email(counter: ToolUsageCounter):
	user_office_email = get_customization('user_office_email_address')
	message = get_media_file_contents('counter_threshold_reached_email.html')
	if user_office_email and message:
		subject = f"Warning threshold reached for {counter.tool.name} {counter.name} counter"
		rendered_message = Template(message).render(Context({'counter': counter}))
		send_mail(subject=subject, content=rendered_message, from_email=user_office_email, to=counter.warning_email, email_category=EmailCategory.SYSTEM)


from datetime import timedelta

import requests
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.shortcuts import render
from django.urls import Resolver404, resolve
from django.utils import timezone
from django.views.decorators.http import require_GET

from NEMO.models import Alert, LandingPageChoice, Reservation, Resource, UsageEvent, User
from NEMO.views.alerts import delete_expired_alerts
from NEMO.views.area_access import able_to_self_log_in_to_area, able_to_self_log_out_of_area
from NEMO.views.notifications import delete_expired_notifications


@login_required
@require_GET
def landing(request):
	user: User = request.user
	delete_expired_alerts()
	delete_expired_notifications()
	usage_events = UsageEvent.objects.filter(operator=user.id, end=None).prefetch_related("tool", "project")
	tools_in_use = [u.tool.tool_or_parent_id() for u in usage_events]
	fifteen_minutes_from_now = timezone.now() + timedelta(minutes=15)
	landing_page_choices = LandingPageChoice.objects.all()
	if request.device == "desktop":
		landing_page_choices = landing_page_choices.exclude(hide_from_desktop_computers=True)
	if request.device == "mobile":
		landing_page_choices = landing_page_choices.exclude(hide_from_mobile_devices=True)
	if not user.is_staff and not user.is_superuser and not user.is_technician:
		landing_page_choices = landing_page_choices.exclude(hide_from_users=True)

	if not settings.ALLOW_CONDITIONAL_URLS:
		# validate all urls
		landing_page_choices = [
			landing_page_choice
			for landing_page_choice in landing_page_choices
			if valid_url_for_landing(landing_page_choice.url)
		]

	upcoming_reservations = Reservation.objects.filter(
		user=user.id, end__gt=timezone.now(), cancelled=False, missed=False, shortened=False
	).exclude(tool_id__in=tools_in_use, start__lte=fifteen_minutes_from_now).exclude(ancestor__shortened=True)
	if user.in_area():
		upcoming_reservations = upcoming_reservations.exclude(
			area=user.area_access_record().area, start__lte=fifteen_minutes_from_now
		)
	upcoming_reservations = upcoming_reservations.order_by("start")[:3]
	dictionary = {
		"now": timezone.now(),
		"alerts": Alert.objects.filter(
			Q(user=None) | Q(user=user), debut_time__lte=timezone.now(), expired=False, deleted=False
		),
		"usage_events": usage_events,
		"upcoming_reservations": upcoming_reservations,
		"disabled_resources": Resource.objects.filter(available=False),
		"landing_page_choices": landing_page_choices,
		"self_log_in": able_to_self_log_in_to_area(request.user),
		"self_log_out": able_to_self_log_out_of_area(request.user),
	}
	return render(request, "landing.html", dictionary)


def valid_url_for_landing(url) -> bool:
	if url.startswith("/"):
		# Internal URL. let's check if it resolves
		try:
			resolve(url)
		except Resolver404:
			return False
	else:
		# External URL, let's check if it exists
		try:
			response = requests.head(url, timeout=0.3)
			if response.status_code > 400:
				return False
		except Exception:
			return False
	return True
